@@ -14,7 +14,6 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 # ---------------------------
-
 """Re-usable functions and private sub functions for common actions.
 
 Timeouts and exceptions are handled in decorators.
@@ -25,12 +24,17 @@ Coding rules:
     3.) All exception and retry related logic should be handle
         in decorators for consistency and usability reasons.
 """
+from __future__ import annotations
+from typing import Optional, Union, Any
+
 import time
 import fnmatch
 from robot.api import logger
+from selenium.webdriver.remote.webelement import WebElement
+from selenium.webdriver.support.select import Select
 from selenium.webdriver.common.action_chains import ActionChains
 from selenium.common.exceptions import WebDriverException, NoSuchElementException, \
-    MoveTargetOutOfBoundsException
+    MoveTargetOutOfBoundsException, ElementNotInteractableException
 from QWeb.internal.exceptions import QWebValueMismatchError, QWebValueError, \
     QWebUnexpectedConditionError, QWebInvalidElementStateError, QWebTimeoutError, \
     QWebTextNotFoundError
@@ -41,42 +45,60 @@ from QWeb.internal.config_defaults import CONFIG
 
 
 @decorators.timeout_decorator_for_actions
-def write(input_element, input_text, timeout, **kwargs):  # pylint: disable=unused-argument
+def write(
+        input_element: WebElement,
+        input_text: str,
+        timeout: int,  # pylint: disable=unused-argument
+        **kwargs: Any) -> None:
     click = util.par2bool(kwargs.get('click', CONFIG['ClickToFocus']))
     if click:
         wd_click(input_element)
     if _ends_with_line_break(input_text):
-        input_text, key = _remove_ending_line_break(input_text)
+        # input_text might be set to None from _remove_ending_line_break(input_text)
+        input_text, key = _remove_ending_line_break(  # type: ignore[assignment]
+            input_text)
         kwargs['key'] = key
     check = util.par2bool(kwargs.get('check', CONFIG['CheckInputValue']))
+    kwargs['shadow_dom'] = CONFIG['ShadowDOM']
     if check is True:
         kwargs['check'] = True
         input_handler.write(input_element, input_text, **kwargs)
         time.sleep(1)
-        compare_input_values(input_element, kwargs.get('expected', input_text), timeout=2)
-        input_element.send_keys(kwargs.get('key', input_handler.line_break_key))
+        compare_input_values(input_element, kwargs.get('expected', input_text), timeout=2, **kwargs)
+        try:
+            input_element.send_keys(kwargs.get('key', input_handler.line_break_key))
+        except ElementNotInteractableException:
+            # this can happen with firefox for shadow dom elements
+            # log, but do not fail the test case as value was written correctly
+            logger.debug("Could not send line break key to input")
     else:
         input_handler.write(input_element, input_text, **kwargs)
     logger.debug('Preferred text: "{}"'.format(input_text))
 
 
 @decorators.timeout_decorator_for_actions
-def compare_input_values(input_element, expected_value, timeout):  # pylint: disable=unused-argument
+def compare_input_values(
+        input_element: WebElement,
+        expected_value: str,
+        timeout: int,  # pylint: disable=unused-argument
+        **kwargs: Any) -> bool:
     try:
-        real_value = util.get_substring(input_value(input_element, timeout="0.5s"))
+        real_value = util.get_substring(input_value(input_element, timeout="0.5s", **kwargs))
     except QWebValueError:
         real_value = ""
     logger.debug('Real value: {}, expected value {}'.format(real_value, expected_value))
-    if fnmatch.fnmatch(real_value.strip(), expected_value):
+    if fnmatch.fnmatch(str(real_value).strip(), expected_value):
         return True
-    raise QWebValueMismatchError('Expected value "{}" didn\'t match to real value "{}"'
-                                 .format(expected_value, real_value))
+    raise QWebValueMismatchError('Expected value "{}" didn\'t match to real value "{}"'.format(
+        expected_value, real_value))
 
 
 @decorators.timeout_decorator_for_actions
-def input_value(input_element, timeout, **kwargs):
+def input_value(input_element: WebElement, timeout: int, **kwargs: Any) -> str:
     blind = util.par2bool(kwargs.get('blind', CONFIG['BlindReturn']))
-    if input_handler.is_editable_text_element(input_element):
+    shadow_dom = CONFIG['ShadowDOM']
+
+    if input_handler.is_editable_text_element(input_element) and not shadow_dom:
         value = input_element.get_attribute('innerText')
     else:
         value = input_element.get_attribute('value')
@@ -88,12 +110,14 @@ def input_value(input_element, timeout, **kwargs):
 
 
 @decorators.timeout_decorator_for_actions
-def scroll(web_element, timeout):  # pylint: disable=unused-argument
+def scroll(web_element: WebElement, timeout: int) -> None:  # pylint: disable=unused-argument
     javascript.execute_javascript('arguments[0].scrollIntoView();', web_element)
 
 
 @decorators.timeout_decorator_for_actions
-def execute_click_and_verify_condition(web_element, text_appear=True, **kwargs):
+def execute_click_and_verify_condition(web_element: WebElement,
+                                       text_appear: bool = True,
+                                       **kwargs: Any) -> bool:
     """Click and optionally verify condition after click.
 
     Accepted kwargs:
@@ -102,14 +126,17 @@ def execute_click_and_verify_condition(web_element, text_appear=True, **kwargs):
         timeout = How long we are trying if element is not enabled or some other error exists
         js = if js parameter exists, try javascript click instead of selenium
     """
-    js = util.par2bool(kwargs.get('js', False))
+    js = True if util.is_safari() else util.par2bool(kwargs.get('js', False))
     dbl_click = util.par2bool(kwargs.get('doubleclick', CONFIG["DoubleClick"]))
     if web_element.is_enabled():
         try:
-            if js:
+            if dbl_click:
+                if js:
+                    js_double_click(web_element)
+                else:
+                    double_click(web_element)
+            elif js:
                 js_click(web_element)
-            elif dbl_click:
-                double_click(web_element)
             else:
                 wd_click(web_element, **kwargs)
                 logger.debug('element clicked')
@@ -120,28 +147,38 @@ def execute_click_and_verify_condition(web_element, text_appear=True, **kwargs):
         if 'text' in kwargs:
             logger.debug('button clicked. Verifying expected condition..')
             try:
-                if text_appearance(kwargs['text'], text_appear=text_appear,
+                if text_appearance(kwargs['text'],
+                                   text_appear=text_appear,
                                    timeout=kwargs.get('interval')):
                     return True
-            except QWebTimeoutError:
+            except QWebTimeoutError as e:
                 logger.debug('timeout err')
-                raise QWebUnexpectedConditionError('Unexpected condition')
+                raise QWebUnexpectedConditionError('Unexpected condition') from e
         return True
     raise QWebInvalidElementStateError('Element is not enabled')
 
 
-def right_click(element):
+def right_click(element: WebElement) -> None:
     driver = browser.get_current_browser()
     ac = ActionChains(driver)
     ac.context_click(element).perform()
 
 
-def js_click(web_element):
+def js_click(web_element: WebElement) -> None:
     javascript.execute_javascript('arguments[0].click()', web_element)
     logger.debug("Js click performed")
 
 
-def double_click(web_element):
+def js_double_click(web_element: WebElement) -> None:
+    js = """var target = arguments[0];
+            var clickEvent = document.createEvent('MouseEvents');
+            clickEvent.initEvent ('dblclick', true, true);
+            target.dispatchEvent(clickEvent);"""
+    javascript.execute_javascript(js, web_element)
+    logger.debug("Js double-click performed")
+
+
+def double_click(web_element: WebElement) -> None:
     driver = browser.get_current_browser()
     ac = ActionChains(driver)
     ac.double_click(web_element)
@@ -149,7 +186,7 @@ def double_click(web_element):
     logger.debug('element double-clicked')
 
 
-def wd_click(web_element, **kwargs):
+def wd_click(web_element: WebElement, **kwargs: Any) -> None:
     try:
         web_element.click()
     except WebDriverException as e:
@@ -160,8 +197,11 @@ def wd_click(web_element, **kwargs):
 
 
 @decorators.timeout_decorator_for_actions
-def checkbox_set(checkbox_element, locator_element, value,
-                 **kwargs):  # pylint: disable=unused-argument
+def checkbox_set(  # pylint: disable=unused-argument
+        checkbox_element: WebElement,
+        locator_element: WebElement,
+        value: bool,
+        **kwargs: Any) -> None:
     if checkbox.is_checked(checkbox_element) != value:
         try:
             checkbox_element.click()
@@ -176,7 +216,11 @@ def checkbox_set(checkbox_element, locator_element, value,
 
 # pylint: disable=too-many-branches
 @decorators.timeout_decorator_for_actions
-def select_option(select, option, unselect=False, **kwargs):  # pylint: disable=unused-argument
+def select_option(
+        select: Select,  # pylint: disable=unused-argument
+        option: str,
+        unselect: bool = False,
+        **kwargs: Any) -> bool:
     """Click and optionally verify condition after click.
 
     Parameters
@@ -199,8 +243,8 @@ def select_option(select, option, unselect=False, **kwargs):  # pylint: disable=
                 else:
                     select.select_by_index(option)
                 return True
-            except TypeError:
-                raise QWebValueMismatchError('Index out of range')
+            except TypeError as te:
+                raise QWebValueMismatchError('Index out of range') from te
     try:
         if unselect:
             select.deselect_by_visible_text(option)
@@ -220,28 +264,30 @@ def select_option(select, option, unselect=False, **kwargs):  # pylint: disable=
                     option_list.append(opt.text)
                     value_list.append(opt.get_attribute('value'))
         if option_list != value_list:
-            raise QWebValueMismatchError('Option "{}" is not in the options list.\n'
-                                         'The list contained these options: {}.\n'
-                                         'The list contained these values: {}.'.format(option,
-                                                                                       option_list,
-                                                                                       value_list))
-        raise QWebValueMismatchError('Option "{}" is not in the options list.\n'
-                                     'The list contained these options: {}.\n'
-                                     .format(option, option_list))
+            raise QWebValueMismatchError(  # pylint: disable=W0707
+                f'Option "{option}" is not in the options list.\n'
+                f'The list contained these options: {option_list}.\n'
+                f'The list contained these values: {value_list}.')
+        raise QWebValueMismatchError(  # pylint: disable=W0707
+            f'Option "{option}" is not in the options list.\n'
+            'The list contained these options: {option_list}.\n')
 
 
 @decorators.timeout_decorator_for_actions
-def is_not_in_dropdown(select, option, **kwargs):
+def is_not_in_dropdown(select: Select, option: str, **kwargs: Any) -> bool:
     """"Verifies that the selected option is not in the dropdown list"""
     option_list = get_select_options(select, **kwargs)
     if option in option_list:
-        raise QWebValueError("Found the value {} from the dropdown menu: {}."
-                             .format(option, option_list))
+        raise QWebValueError("Found the value {} from the dropdown menu: {}.".format(
+            option, option_list))
     return True
 
 
 @decorators.timeout_decorator_for_actions
-def get_selected_value(select, expected=None, **kwargs):  # pylint: disable=unused-argument
+def get_selected_value(
+        select: Select,  # pylint: disable=unused-argument
+        expected: Optional[str] = None,
+        **kwargs: Any) -> Union[bool, str]:
     """Get or verify selected value.
 
     Parameters
@@ -259,13 +305,16 @@ def get_selected_value(select, expected=None, **kwargs):  # pylint: disable=unus
     if expected:
         if expected in selected:
             return True
-        raise QWebValueMismatchError('Expected value "{}" didn\'t match to real value "{}".'
-                                     .format(expected, txt_selected))
+        raise QWebValueMismatchError('Expected value "{}" didn\'t match to real value "{}".'.format(
+            expected, txt_selected))
     return txt_selected
 
 
 @decorators.timeout_decorator_for_actions
-def get_select_options(select, expected=None, **kwargs):  # pylint: disable=unused-argument
+def get_select_options(
+        select: Select,  # pylint: disable=unused-argument
+        expected: Optional[str] = None,
+        **kwargs: Any) -> Union[bool, list[str]]:
     options = select.options
     if expected:
         for option in options:
@@ -282,11 +331,16 @@ def get_select_options(select, expected=None, **kwargs):  # pylint: disable=unus
 
 
 @decorators.timeout_decorator_for_actions
-def hover_to(web_element, timeout=0):  # pylint: disable=unused-argument
+def hover_to(web_element: WebElement, timeout: int = 0) -> None:  # pylint: disable=unused-argument
     driver = browser.get_current_browser()
-    # firefox specific fix
-    if driver.capabilities['browserName'].lower() in browser.firefox.NAMES:
+    # firefox & safari specific fix
+    needs_js_scroll = [
+        x for x in [browser.firefox.NAMES, browser.safari.NAMES]
+        if driver.capabilities['browserName'].lower() in x
+    ]
+    if needs_js_scroll:
         # use javascript to scroll
+        logger.debug("Needs javascript to scoll")
         driver.execute_script("arguments[0].scrollIntoView(true);", web_element)
     try:
         hover = ActionChains(driver).move_to_element(web_element)
@@ -297,7 +351,7 @@ def hover_to(web_element, timeout=0):  # pylint: disable=unused-argument
 
 
 @decorators.timeout_decorator_for_actions
-def text_appearance(text, **kwargs):
+def text_appearance(text: str, **kwargs: Any) -> bool:
     """ Sub for retry click.
 
     Works as keywords is_text and is_no_text. Returns True if
@@ -305,12 +359,11 @@ def text_appearance(text, **kwargs):
     for decorator to handle if condition is not expected (False).
     """
     try:
-        element = internal_text.get_element_by_locator_text(
-            text, allow_non_existent=True, **kwargs)
-    except QWebTimeoutError:
+        element = internal_text.get_element_by_locator_text(text, allow_non_existent=True, **kwargs)
+    except QWebTimeoutError as te:
         if kwargs['text_appear'] is False:
             return True
-        raise QWebValueMismatchError('return value should be true')
+        raise QWebValueMismatchError('return value should be true') from te
     try:
         if element and kwargs['text_appear'] is True:
             return True
@@ -322,52 +375,53 @@ def text_appearance(text, **kwargs):
 
 
 @decorators.timeout_decorator_for_actions
-def get_element_text(web_element, expected=None, timeout=0):  # pylint: disable=unused-argument
+def get_element_text(web_element: WebElement, expected=None, timeout: int = 0) -> Union[bool, str]:  # pylint: disable=unused-argument
     real_text = web_element.text.strip()
     if expected is not None:
         try:
             return _compare_texts(real_text, expected.strip(), timeout)
-        except QWebValueMismatchError:
-            raise QWebValueError('Expected {}, found {}'.format(expected, real_text))
+        except QWebValueMismatchError as e:
+            raise QWebValueError('Expected {}, found {}'.format(expected, real_text)) from e
     if real_text is not None:
         return real_text
     raise QWebValueMismatchError('Text not found')
 
 
-def _compare_texts(text_to_compare, expected, timeout):  # pylint: disable=unused-argument
+def _compare_texts(text_to_compare: str, expected: str, timeout: int) -> bool:  # pylint: disable=unused-argument
     if fnmatch.fnmatch(text_to_compare, expected) is False:
-        raise QWebValueMismatchError(
-            'Expected {0}, found {1}'.format(expected, text_to_compare))
+        raise QWebValueMismatchError('Expected {0}, found {1}'.format(expected, text_to_compare))
     return True
 
 
-def _ends_with_line_break(input_text):
+def _ends_with_line_break(input_text: str) -> bool:
     line_break = ('\n', '\ue007', '\t', '\ue004')
     return input_text.endswith(line_break)
 
 
-def _remove_ending_line_break(input_text):
+def _remove_ending_line_break(input_text: str) -> tuple[Optional[str], Optional[str]]:
     enter_key = ('\n', '\ue007')
     tab_key = ('\t', '\ue004')
     if input_text.endswith(enter_key):
-        input_text = input_text.replace('\n', '\ue007').rsplit('\ue007', 1)
-        return input_text[0], '\ue007'
+        input_list = input_text.replace('\n', '\ue007').rsplit('\ue007', 1)
+        return input_list[0], '\ue007'
     if input_text.endswith(tab_key):
-        input_text = input_text.replace('\t', '\ue004').rsplit('\ue004', 1)
-        return input_text[0], '\ue004'
+        input_list = input_text.replace('\t', '\ue004').rsplit('\ue004', 1)
+        return input_list[0], '\ue004'
     return None, None
 
 
-def _contains_enter(input_text):
+def _contains_enter(input_text: str) -> bool:
     return bool('\n' in input_text or '\ue007' in input_text)
 
 
-def _contains_tab(input_text):
+def _contains_tab(input_text: str) -> bool:
     return bool('\t' in input_text or '\ue004' in input_text)
 
 
-def scroll_first_scrollable_parent_element(locator, anchor, text_to_find, scroll_length,
-                                           slow_mode, timeout):
+def scroll_first_scrollable_parent_element(locator: str, anchor: str, text_to_find: str,
+                                           scroll_length: Optional[Union[int, str]],
+                                           slow_mode: bool, timeout: Union[int, float,
+                                                                           str]) -> None:
     visible = None
     js_get_parent_element = """
         function getScrollParent(node) {
@@ -391,19 +445,21 @@ def scroll_first_scrollable_parent_element(locator, anchor, text_to_find, scroll
     old_pos = None
     start = time.time()
     if not slow_mode:
-        while not visible and old_pos != current_pos and time.time() < timeout + start:
+        while not visible and old_pos != current_pos and time.time() < float(timeout) + start:
             old_pos = javascript.execute_javascript(js_element_position, scrollable_element)
             javascript.execute_javascript(js_element_scroll, scrollable_element)
             time.sleep(.5)
             current_pos = javascript.execute_javascript(js_element_position, scrollable_element)
             visible = internal_text.get_element_by_locator_text(text_to_find,
                                                                 allow_non_existent=True)
-            logger.info('Old pos: {}\nNew pos: {}\nVisible: {}'
-                        .format(old_pos, current_pos, visible), also_console=True)
+            logger.info('Old pos: {}\nNew pos: {}\nVisible: {}'.format(
+                old_pos, current_pos, visible),
+                        also_console=True)
     else:
         logger.info('\nSlow mode is on, execution will only stop if the text "{}" is found or if '
-                    'the timeout is reached.'.format(text_to_find), also_console=True)
-        while not visible and time.time() < timeout + start:
+                    'the timeout is reached.'.format(text_to_find),
+                    also_console=True)
+        while not visible and time.time() < float(timeout) + start:
             javascript.execute_javascript(js_element_scroll, scrollable_element)
             time.sleep(.5)
             visible = internal_text.get_element_by_locator_text(text_to_find,
@@ -414,7 +470,8 @@ def scroll_first_scrollable_parent_element(locator, anchor, text_to_find, scroll
     raise QWebTextNotFoundError('Text {} not found.'.format(text_to_find))
 
 
-def scroll_dynamic_web_page(text_to_find, scroll_length, slow_mode, timeout):
+def scroll_dynamic_web_page(text_to_find: str, scroll_length: Optional[Union[int, str]],
+                            slow_mode: bool, timeout: Union[int, float, str]) -> bool:
     visible = None
     js_browser_height = "return window.innerHeight"
     height = javascript.execute_javascript(js_browser_height)  # Length of one scroll
@@ -424,19 +481,21 @@ def scroll_dynamic_web_page(text_to_find, scroll_length, slow_mode, timeout):
     old_pos = None
     start = time.time()
     if not slow_mode:
-        while not visible and old_pos != current_pos and time.time() < timeout + start:
+        while not visible and old_pos != current_pos and time.time() < float(timeout) + start:
             old_pos = javascript.execute_javascript(js_current_pos)
             javascript.execute_javascript(js_scroll)
             time.sleep(.5)
             current_pos = javascript.execute_javascript(js_current_pos)
             visible = internal_text.get_element_by_locator_text(text_to_find,
                                                                 allow_non_existent=True)
-            logger.info('\nOld pos: {}\nCurrent pos: {}\nVisible: {}'
-                        .format(old_pos, current_pos, visible), also_console=True)
+            logger.info('\nOld pos: {}\nCurrent pos: {}\nVisible: {}'.format(
+                old_pos, current_pos, visible),
+                        also_console=True)
     else:
         logger.info('\nSlow mode is on, execution will only stop if the text "{}" is found or if '
-                    'the timeout is reached.'.format(text_to_find), also_console=True)
-        while not visible and time.time() < timeout + start:
+                    'the timeout is reached.'.format(text_to_find),
+                    also_console=True)
+        while not visible and time.time() < float(timeout) + start:
             javascript.execute_javascript(js_scroll)
             time.sleep(.5)
             visible = internal_text.get_element_by_locator_text(text_to_find,
@@ -445,5 +504,5 @@ def scroll_dynamic_web_page(text_to_find, scroll_length, slow_mode, timeout):
 
     if visible:
         return True
-    raise QWebTextNotFoundError('Could not find text "{}" after scrolling for {} pixels.'
-                                .format(text_to_find, current_pos))
+    raise QWebTextNotFoundError('Could not find text "{}" after scrolling for {} pixels.'.format(
+        text_to_find, current_pos))
