@@ -14,14 +14,20 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 # ---------------------------
+from __future__ import annotations
+from numpy import ndarray
 
 import cv2
 import numpy as np
 import math
+import os
 from pathlib import Path
 from QWeb.internal import frame, download
 from QWeb.internal.meas import MEAS
-from QWeb.internal.screenshot import save_screenshot
+from QWeb.internal.screenshot import save_screenshot, log_screenshot_file, SCREEN_SHOT_DIR_NAME
+from QWeb.internal.config_defaults import CONFIG
+from robot.libraries.BuiltIn import BuiltIn
+from uuid import uuid4
 
 
 class QIcon:
@@ -31,7 +37,7 @@ class QIcon:
         pass
 
     @staticmethod
-    def _get_image_size(image):
+    def _get_image_size(image: ndarray) -> tuple[int, int]:
         """Returns image width and height
         """
         # pylint: disable=no-self-use
@@ -39,7 +45,7 @@ class QIcon:
         return screen_w, screen_h
 
     @staticmethod
-    def _get_scale_ratios(template_res_w, device_res_w):
+    def _get_scale_ratios(template_res_w: int, device_res_w: int) -> list[float]:
         """Get a list of different scale ratios to scale template image
         for different device resolutions. They should be ordered so
         that more common resolution scalings are at the top of the list.
@@ -57,8 +63,9 @@ class QIcon:
             Resolution of the device we are testing
         """
 
-        scale_ratios = [1.00, 0.75, 0.50, 0.86, 0.78,
-                        0.58, 1.33, 0.67, 1.15]
+        scale_ratios = [
+            1.00, 0.75, 0.50, 0.86, 0.78, 0.58, 1.33, 0.67, 1.15, 2.00, 1.50, 1.25, 1.75
+        ]
 
         if device_res_w <= 0 or template_res_w <= 0:
             raise ValueError("Device resolution {} or template resolution {}"
@@ -66,10 +73,11 @@ class QIcon:
 
         if not round(device_res_w / template_res_w, 2) in scale_ratios:
             scale_ratios.insert(0, round(device_res_w / template_res_w, 2))
+
         return scale_ratios
 
     @staticmethod
-    def _get_image_pyramid(image_obj, level):
+    def _get_image_pyramid(image_obj: ndarray, level: int) -> list[tuple[ndarray, float]]:
         """
         Returns a list of up- and downsampled images of image_obj.
         Each sampling goes 10% up and down in size
@@ -101,18 +109,28 @@ class QIcon:
                     #                 "Scale coefficient: {}."
                     #                 .format(scale))
                 if scale < 1.0:
-                    image_levels.append((cv2.resize(image_obj, None, fx=scale, fy=scale,
-                                                    interpolation=cv2.INTER_LINEAR),
-                                         scale))
+                    image_levels.append((cv2.resize(image_obj,
+                                                    None,
+                                                    fx=scale,
+                                                    fy=scale,
+                                                    interpolation=cv2.INTER_LINEAR), scale))
                 elif scale == 1.0:
                     image_levels.append((image_obj, scale))
                 else:  # scale > 1.0
-                    image_levels.append((cv2.resize(image_obj, None, fx=scale, fy=scale,
-                                                    interpolation=cv2.INTER_LINEAR),
-                                         scale))
+                    image_levels.append((cv2.resize(image_obj,
+                                                    None,
+                                                    fx=scale,
+                                                    fy=scale,
+                                                    interpolation=cv2.INTER_LINEAR), scale))
         return image_levels
 
-    def get_template_locations(self, image_obj, template, threshold, level=0):  # pylint: disable=unused-argument
+    def get_template_locations(
+            self,
+            image_obj: ndarray,
+            template: ndarray,
+            threshold: float,
+            level: int = 0  # pylint: disable=unused-argument
+    ) -> list[tuple[int, int]]:
         """Returns a list of template locations. Uses image pyramid.
         _image_obj_ = Image as uint8 NumPy array in RGB color space.
         _threshold_ = Threshold value for template matching, float.
@@ -134,8 +152,8 @@ class QIcon:
 
         if height > img_height or width > img_width:
             raise ValueError("Image Size Error: Template image is larger than source image. "
-                             "Template size: {}x{}, source image size: {}x{}."
-                             .format(width, height, img_width, img_height))
+                             "Template size: {}x{}, source image size: {}x{}.".format(
+                                 width, height, img_width, img_height))
         print("*DEBUG* Resampling by 16 levels for needle, 0 levels for haystack")
 
         # temp = image_obj
@@ -154,7 +172,7 @@ class QIcon:
         best_highest_max_val_loc = (-1, -1)
         points = []
         best_scale = 0.0
-        best_matched_image = None
+        best_matched_image: ndarray  # = None
 
         print("*DEBUG* Different resamplings used: " + str(len(template_levels)))
         # for template_level in template_levels:
@@ -172,10 +190,8 @@ class QIcon:
             # print("did matchTemplate, res:\n")
             # print(res)
             MEAS.start("EXTRACT POINTS TIME")
-            current_points, highest_max_val, highest_max_val_loc = self._extract_points(h,
-                                                                                        res,
-                                                                                        threshold,
-                                                                                        w)
+            current_points, highest_max_val, highest_max_val_loc = self._extract_points(
+                h, res, threshold, w)
             MEAS.stop()
             # print("did extract_points, points:\n")
             # print(current_points)
@@ -194,31 +210,25 @@ class QIcon:
         else:
             print("*INFO* Template image not found. "
                   "Closest match threshold value {} found at {}, {}. "
-                  "Threshold used: {}"
-                  .format(best_highest_max_val,
-                          best_highest_max_val_loc[0],
-                          best_highest_max_val_loc[1],
-                          threshold))
+                  "Threshold used: {}".format(best_highest_max_val, best_highest_max_val_loc[0],
+                                              best_highest_max_val_loc[1], threshold))
         print("*INFO* Closest match threshold value {} found at {}, {}.\n"
               "Threshold used: {}\n"
-              "Scale with best result: {}"
-              .format(best_highest_max_val,
-                      best_highest_max_val_loc[0],
-                      best_highest_max_val_loc[1],
-                      threshold,
-                      best_scale))
-        self._log_matched_image(color_image_obj, color_template,
-                                best_matched_image, best_highest_max_val_loc, best_scale)
+              "Scale with best result: {}".format(best_highest_max_val, best_highest_max_val_loc[0],
+                                                  best_highest_max_val_loc[1], threshold,
+                                                  best_scale))
+        self._log_matched_image(color_image_obj, color_template, best_matched_image,
+                                best_highest_max_val_loc, best_scale)
         print("*DEBUG {}".format(points))
         return points
 
     def image_location(self,
-                       needle,
-                       haystack,
-                       tolerance=0.95,
-                       draw=1,
-                       template_res_w=1440,
-                       device_res_w=1080):
+                       needle: str,
+                       haystack: str,
+                       tolerance: float = 0.95,
+                       draw: int = 1,
+                       template_res_w: int = 1440,
+                       device_res_w: int = 1080) -> tuple[int, int]:
         """Locate an image (needle) within an bigger image (haystack). Tolarance
         is pixel tolerance, i.e. 1.0 = all pixels are correct, 0.5 = 50% of the pixels
         are correct. If we know the original resolution, from which the template
@@ -247,34 +257,42 @@ class QIcon:
 
         best_highest_max_val = 0.0
         best_highest_max_val_loc = (-1, -1)
-        best_scale_ratio = None
-        best_matched_image = None
+        best_scale_ratio: float  # = None
+        best_matched_image: ndarray  # = None
 
         print("*DEBUG* Resampling loop Starts")
         for scale_ratio in scale_ratios:
-            if scale_ratio > 1.0:
-                interpolation_method = cv2.INTER_LINEAR
-            else:
-                interpolation_method = cv2.INTER_AREA
+            interpolation_method = cv2.INTER_LINEAR if scale_ratio > 1.0 else cv2.INTER_AREA
 
             print(("*DEBUG* resize starts: for scale {}".format(scale_ratio)))
 
             if math.isclose(scale_ratio, 1.0, rel_tol=0.03):
                 scaled_img_template = template
             else:
-                scaled_img_template = cv2.resize(template, None, fx=scale_ratio,
-                                                 fy=scale_ratio, interpolation=interpolation_method)
+                scaled_img_template = cv2.resize(template,
+                                                 None,
+                                                 fx=scale_ratio,
+                                                 fy=scale_ratio,
+                                                 interpolation=interpolation_method)
             print("*DEBUG* matchTemplate Starts:")
 
             res = cv2.matchTemplate(image_gray, scaled_img_template, cv2.TM_CCOEFF_NORMED)
 
             ratio = device_res_w / hay_w
+
+            if CONFIG.get_value("RetinaDisplay"):
+                ratio = ratio * 2
+            elif ratio < 1.1:
+                ratio = 1.0
+
             print("*DEBUG* _extract_points Starts:")
-            _current_points, highest_max_val, highest_max_val_loc = self._extract_points(height,
-                                                                                         res,
-                                                                                         tolerance,
-                                                                                         width,
-                                                                                         ratio)
+            _current_points, highest_max_val, highest_max_val_loc = \
+                self._extract_points(height * scale_ratio,
+                                     res,
+                                     tolerance,
+                                     width * scale_ratio,
+                                     ratio)
+
             if highest_max_val > best_highest_max_val:
                 best_highest_max_val = highest_max_val
                 best_highest_max_val_loc = highest_max_val_loc
@@ -283,26 +301,25 @@ class QIcon:
                 best_matched_image = scaled_img_template
                 print("*DEBUG* Current best match location: {},\n"
                       "max_value: {},\n"
-                      "scale_ratio: {}".format(best_highest_max_val_loc,
-                                               best_highest_max_val,
+                      "scale_ratio: {}".format(best_highest_max_val_loc, best_highest_max_val,
                                                best_scale_ratio))
 
                 if best_highest_max_val > tolerance:
                     if draw == 1:
                         loc = np.where(res >= tolerance)
+
                         for pt in zip(*loc[::-1]):
-                            cv2.rectangle(image, pt,
-                                          (pt[0] + width, pt[1] + height),
-                                          (0, 0, 255), 2)
+                            cv2.rectangle(image, pt, (pt[0] + width, pt[1] + height), (0, 0, 255),
+                                          2)
                         cv2.imwrite('temp_matched_area.png', image)
                     break
 
         print("*DEBUG* Ready to return points:")
         print("*DEBUG* Best match location: {}, best correlation value: {}, "
-              "best scale ratio: {}".format(best_highest_max_val_loc,
-                                            best_highest_max_val, best_scale_ratio))
-        self._log_matched_image(image, template, best_matched_image,
-                                best_highest_max_val_loc, best_scale_ratio)
+              "best scale ratio: {}".format(best_highest_max_val_loc, best_highest_max_val,
+                                            best_scale_ratio))
+        self._log_matched_image(image, template, best_matched_image, best_highest_max_val_loc,
+                                best_scale_ratio)
 
         if best_highest_max_val >= tolerance:
             return best_highest_max_val_loc
@@ -310,7 +327,12 @@ class QIcon:
         return -1, -1
 
     @staticmethod
-    def _extract_points(height, res, threshold, width, coordinate_ratio=1.0):
+    def _extract_points(height: int,
+                        res: ndarray,
+                        threshold: float,
+                        width: int,
+                        coordinate_ratio: float = 1.0
+                        ) -> tuple[list[tuple[int, int]], float, tuple[int, int]]:
         """Extracts match locations from a result matrix. This goes through the best
         match locations in the result and while the correlation value of a location
         is over the threshold, it adds the location to the list of best locations.
@@ -351,12 +373,13 @@ class QIcon:
               "Coordinate ratio was {}\n"
               "Highest max value was {}\n"
               "Highest max value location was {}\n"
-              "All points: {}".format(coordinate_ratio, highest_max_val,
-                                      highest_max_val_loc, points))
+              "All points: {}".format(coordinate_ratio, highest_max_val, highest_max_val_loc,
+                                      points))
         return points, highest_max_val, highest_max_val_loc
 
     @staticmethod
-    def _log_matched_image(haystack, needle, scaled_needle, loc, best_scale):
+    def _log_matched_image(haystack: ndarray, needle: ndarray, scaled_needle: ndarray,
+                           loc: tuple[int, int], best_scale: float) -> None:
         """Draw a composite image with the needle image, the haystack image,
         the scaled needle that matches the best and show where in haystack
         the best match is. This is best used in debugging, but it could be
@@ -369,29 +392,29 @@ class QIcon:
         h2, w2 = haystack.shape[:2]
         max_left_w = max(w1, ws)
         cv2.rectangle(haystack,
-                      (loc[0] - int(w1 / 2 * best_scale),
-                       loc[1] - int(h1 / 2 * best_scale)),
-                      (loc[0] + int(w1 / 2 * best_scale),
-                       loc[1] + int(h1 / 2 * best_scale)),
+                      (loc[0] - int(w1 / 2 * best_scale), loc[1] - int(h1 / 2 * best_scale)),
+                      (loc[0] + int(w1 / 2 * best_scale), loc[1] + int(h1 / 2 * best_scale)),
                       (0, 0, 255), 2)
         result = np.zeros((max(h2, h1), w2 + max_left_w, 3), np.uint8)
         result[:h1, :w1, :3] = needle
         result[h1:h1 + hs, :ws, :3] = scaled_needle
         result[:h2, max_left_w:max_left_w + w2, :3] = haystack
 
-        cv2.line(result,
-                 (ws, h1),
-                 (loc[0] + max_left_w + int(ws / 2), loc[1] - int(hs / 2)),
+        cv2.line(result, (ws, h1), (loc[0] + max_left_w + int(ws / 2), loc[1] - int(hs / 2)),
                  (0, 0, 255), 2)
-        cv2.line(result,
-                 (0, h1 + hs),
-                 (loc[0] + max_left_w - int(ws / 2), loc[1] + int(hs / 2)),
+        cv2.line(result, (0, h1 + hs), (loc[0] + max_left_w - int(ws / 2), loc[1] + int(hs / 2)),
                  (0, 0, 255), 2)
 
-        cv2.imwrite('temp_matched_image.png', result)
+        if CONFIG.get_value("LogMatchedIcons"):
+            output = BuiltIn().get_variable_value('${OUTPUT DIR}')
+            filename = "temp_matched_image-{}".format(uuid4()) + ".png"
+            filepath = os.path.join(output, SCREEN_SHOT_DIR_NAME, filename)
+            cv2.imwrite(filepath, result)
+            log_screenshot_file(filepath)
 
 
-def image_recognition(image_path, template_res_w, browser_res_w, pyautog):
+def image_recognition(image_path: str, template_res_w: int, browser_res_w: int,
+                      pyautog: bool) -> tuple[int, int]:
     """Return icon's coordinates."""
     image_rec = QIcon()
     frame.wait_page_loaded()
@@ -403,7 +426,7 @@ def image_recognition(image_path, template_res_w, browser_res_w, pyautog):
     return x, y
 
 
-def get_full_image_path(icon):
+def get_full_image_path(icon: str) -> Path:
     """Return image's full path."""
     if icon.endswith('.png') or icon.endswith('.jpg'):
         full_path = download.get_path(icon)
