@@ -14,17 +14,24 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 # ---------------------------
-
 """Keywords for general elements that are retrieved using XPaths."""
+from __future__ import annotations
+from typing import Union, Optional
+from selenium.webdriver.remote.webelement import WebElement
 
+from robot.api.deco import keyword
 from QWeb.internal.exceptions import QWebValueError, QWebElementNotFoundError
 from QWeb.internal import element, decorators, actions, text, input_,\
     dropdown, checkbox
-from selenium.webdriver.remote.webelement import WebElement
 
 
+@keyword(tags=["Interaction"])
 @decorators.timeout_decorator
-def click_element(xpath, timeout=0, js=False, index=1, **kwargs):
+def click_element(xpath: Union[WebElement, str],
+                  timeout: Union[int, float, str] = 0,
+                  js: bool = False,
+                  index: int = 1,
+                  **kwargs) -> None:
     r"""Click element specified by xpath.
 
     Examples
@@ -82,16 +89,21 @@ def click_element(xpath, timeout=0, js=False, index=1, **kwargs):
         index = int(index) - 1
         kwargs['element_kw'] = True
         if 'tag' in kwargs:
-            web_element = element.get_elements_by_attributes(
-                kwargs.get('tag'), xpath, **kwargs)[index]
+            web_element = element.get_elements_by_attributes(kwargs.get('tag'), xpath,
+                                                             **kwargs)[index]
         else:
             web_element = element.get_unique_element_by_xpath(xpath)
     if actions.execute_click_and_verify_condition(web_element, timeout=timeout, js=js, **kwargs):
         return
 
 
+@keyword(tags=["Interaction"])
 @decorators.timeout_decorator
-def right_click(xpath, timeout=0, index=1, **kwargs):  # pylint: disable=unused-argument
+def right_click(
+        xpath: str,
+        timeout: Union[int, float, str] = 0,  # pylint: disable=unused-argument
+        index: int = 1,
+        **kwargs) -> None:
     r"""Right clicks the element.
 
     Examples
@@ -126,16 +138,19 @@ def right_click(xpath, timeout=0, index=1, **kwargs):  # pylint: disable=unused-
     index = int(index) - 1
     kwargs['element_kw'] = True
     if 'tag' in kwargs:
-        web_element = element.get_elements_by_attributes(
-            kwargs.get('tag'), xpath, **kwargs)[index]
+        web_element = element.get_elements_by_attributes(kwargs.get('tag'), xpath, **kwargs)[index]
     else:
         web_element = element.get_unique_element_by_xpath(xpath)
-    if actions.right_click(web_element):
-        return
+    actions.right_click(web_element)
 
 
+@keyword(tags=["Interaction"])
 @decorators.timeout_decorator
-def hover_element(xpath, timeout=0, index=1, **kwargs):  # pylint: disable=unused-argument
+def hover_element(
+        xpath: Union[WebElement, str],
+        timeout: Union[int, float, str] = 0,  # pylint: disable=unused-argument
+        index: int = 1,
+        **kwargs) -> None:
     r"""Hover the element specified by the xpath selector.
 
     Examples
@@ -174,16 +189,20 @@ def hover_element(xpath, timeout=0, index=1, **kwargs):  # pylint: disable=unuse
         index = int(index) - 1
         kwargs['element_kw'] = True
         if 'tag' in kwargs:
-            web_element = element.get_elements_by_attributes(
-                kwargs.get('tag'), xpath, **kwargs)[index]
+            web_element = element.get_elements_by_attributes(kwargs.get('tag'), xpath,
+                                                             **kwargs)[index]
         else:
             web_element = element.get_unique_element_by_xpath(xpath)
 
     actions.hover_to(web_element, timeout=timeout)
 
 
+@keyword(tags=["Getters"])
 @decorators.timeout_decorator
-def get_element_count(locator, timeout=0, **kwargs):  # pylint: disable=unused-argument
+def get_element_count(
+        locator: str,
+        timeout: Union[int, float, str] = 0,  # pylint: disable=unused-argument
+        **kwargs) -> int:
     r"""Get count of appearances for certain web element.
 
     Keyword waits until timeout has passed. If timeout is not specified, it
@@ -214,8 +233,7 @@ def get_element_count(locator, timeout=0, **kwargs):  # pylint: disable=unused-a
     """
     kwargs['element_kw'] = True
     if 'tag' in kwargs:
-        web_elements = element.get_elements_by_attributes(
-            kwargs.get('tag'), locator, **kwargs)
+        web_elements = element.get_elements_by_attributes(kwargs.get('tag'), locator, **kwargs)
     else:
         web_elements = element.get_webelements(locator, **kwargs)
     if web_elements:
@@ -223,7 +241,12 @@ def get_element_count(locator, timeout=0, **kwargs):  # pylint: disable=unused-a
     raise QWebElementNotFoundError('Webelements not found')
 
 
-def is_element(xpath, timeout='0.1s', index=1, **kwargs):  # pylint: disable=unused-argument
+@keyword(tags=["Verification"])
+def is_element(
+        xpath: str,
+        timeout: Union[int, float, str] = '0.1s',
+        index: int = 1,  # pylint: disable=unused-argument
+        **kwargs) -> bool:
     r"""Return True if element is visible.
 
     Examples
@@ -266,8 +289,9 @@ def is_element(xpath, timeout='0.1s', index=1, **kwargs):  # pylint: disable=unu
         return False
 
 
+@keyword(tags=["Verification"])
 @decorators.timeout_decorator
-def verify_element(xpath, timeout=0, **kwargs):  # pylint: disable=unused-argument
+def verify_element(xpath: str, timeout: Union[int, float, str] = 0, **kwargs) -> None:  # pylint: disable=unused-argument
     r"""Verify that element can be found on the page and it is visible.
 
     Examples
@@ -287,7 +311,7 @@ def verify_element(xpath, timeout=0, **kwargs):  # pylint: disable=unused-argume
     Parameters
     ----------
     xpath : str
-        Xpath expression without xpath= prefix. The equal sign "=" must be escaped with a "\".
+        Xpath expression without xpath= prefix. The equal sign "=" must be escaped with a "\\".
     timeout : str | int
         Timeout for finding the element. If the element is not visible after
         given timeout, error is raised. The str is converted to integer
@@ -315,14 +339,18 @@ def verify_element(xpath, timeout=0, **kwargs):  # pylint: disable=unused-argume
         web_elements = element.get_visible_elements_from_elements(
             element.get_elements_by_attributes(kwargs.get('tag'), xpath, **kwargs))
     else:
-        web_elements = element.get_webelements(xpath)
+        web_elements = element.get_webelements(xpath, **kwargs)
     if web_elements:
         return
     raise QWebElementNotFoundError('No matching element found')
 
 
+@keyword(tags=["Verification"])
 @decorators.timeout_decorator
-def verify_no_element(xpath, timeout=0, **kwargs):  # pylint: disable=unused-argument
+def verify_no_element(
+        xpath: str,
+        timeout: Union[int, float, str] = 0,  # pylint: disable=unused-argument
+        **kwargs) -> None:
     r"""Wait element can not be found on the page.
 
     Examples
@@ -342,7 +370,7 @@ def verify_no_element(xpath, timeout=0, **kwargs):  # pylint: disable=unused-arg
     Parameters
     ----------
     xpath : str
-        Xpath expression without xpath= prefix. The equal sign "=" must be escaped with a "\".
+        Xpath expression without xpath= prefix. The equal sign "=" must be escaped with a "\\".
     timeout : str | int
         Timeout for the element to disappear. If the element is visible after
         given timeout, error is raised. The str is converted to integer
@@ -367,19 +395,22 @@ def verify_no_element(xpath, timeout=0, **kwargs):  # pylint: disable=unused-arg
     kwargs['element_kw'] = True
     if 'tag' in kwargs:
         web_elements = element.get_visible_elements_from_elements(
-            element.get_elements_by_attributes(
-                kwargs.get('tag'), xpath, **kwargs))
+            element.get_elements_by_attributes(kwargs.get('tag'), xpath, **kwargs))
     else:
         web_elements = element.get_webelements(xpath)
     if not web_elements:
         return
-    raise QWebValueError(
-        'Page contained element with XPath "{}" after timeout'.format(xpath))
+    raise QWebValueError('Page contained element with XPath "{}" after timeout'.format(xpath))
 
 
+@keyword(tags=["Getters"])
 @decorators.timeout_decorator
-def get_webelement(locator, anchor='1', element_type=None, timeout=0, **kwargs):
-    r"""Get Webelement using any Paceword -stylish locator.
+def get_webelement(locator: str,
+                   anchor: str = '1',
+                   element_type: Optional[str] = None,
+                   timeout: Union[int, float, str] = 0,
+                   **kwargs) -> Union[WebElement, list[WebElement]]:
+    r"""Get Webelement using any Qword -stylish locator.
 
     Examples
     --------
@@ -410,19 +441,25 @@ def get_webelement(locator, anchor='1', element_type=None, timeout=0, **kwargs):
         ${elem}      GetWebelement          Country   element_type=dropdown
         ${elem}      GetWebelement          Gender    element_type=checkbox
 
+    Any element using css selectors:
+
+    .. code-block:: robotframework
+
+        ${elem}      GetWebelement          input[type=button]:nth-child(10)    element_type=css
+
     All flags are available for using (timeout, anchor, index, visibility, parent, child etc.).
-    in same way as you are using those with Pacewords like ClickText/Item, TypeText, Dropdown etc.
+    in same way as you are using those with Qwords like ClickText/Item, TypeText, Dropdown etc.
 
     Parameters
     ----------
     locator : str
         Visible text, attribute value or Xpath expression with or without xpath= prefix.
-        The equal sign "=" must be escaped with a "\".
+        The equal sign "=" must be escaped with a "\\".
     anchor : int
         Used when element_type is defined. Default=1 (first match)
     element_type : string
         Define element type/preferred searching method
-        (available types: text, input, checkbox, item, dropdown).
+        (available types: text, input, checkbox, item, dropdown or css).
     timeout : int
         How long we wait element to appear. Default=10 sec
     kwargs :
@@ -439,15 +476,20 @@ def get_webelement(locator, anchor='1', element_type=None, timeout=0, **kwargs):
     kwargs['timeout'] = timeout
     if element_type:
         if element_type.lower() == 'text':
-            return text.get_element_by_locator_text(locator, anchor, **kwargs)
+            web_elements = text.get_element_by_locator_text(locator, anchor, **kwargs)
         if element_type.lower() == 'item':
-            return text.get_item_using_anchor(locator, anchor, **kwargs)
+            web_elements = text.get_item_using_anchor(locator, anchor, **kwargs)
         if element_type.lower() == "dropdown":
-            return dropdown.get_dd_elements_from_all_documents(locator, anchor, **kwargs)
+            web_elements = dropdown.get_dd_elements_from_all_documents(locator, anchor, **kwargs)
         if element_type.lower() == "input":
-            return input_.get_input_elements_from_all_documents(locator, anchor, **kwargs)
+            web_elements = input_.get_input_elements_from_all_documents(locator, anchor, **kwargs)
         if element_type.lower() == "checkbox":
-            return checkbox.get_checkbox_elements_from_all_documents(locator, anchor, **kwargs)
+            web_elements = checkbox.get_checkbox_elements_from_all_documents(
+                locator, anchor, **kwargs)[0]
+        if element_type.lower() == "css":
+            web_elements = element.get_webelement_by_css(locator, **kwargs)
+        return web_elements
+
     kwargs['element_kw'] = True
     if 'tag' in kwargs:
         web_elements = element.get_visible_elements_from_elements(
@@ -459,8 +501,14 @@ def get_webelement(locator, anchor='1', element_type=None, timeout=0, **kwargs):
     raise QWebElementNotFoundError('No matching element found')
 
 
+@keyword(tags=["Getters"])
 @decorators.timeout_decorator
-def get_attribute(locator, attribute, anchor='1', element_type=None, timeout=0, **kwargs):
+def get_attribute(locator: str,
+                  attribute: str,
+                  anchor: str = '1',
+                  element_type: Optional[str] = None,
+                  timeout: Union[int, float, str] = 0,
+                  **kwargs) -> str:
     r"""Get attribute value of an element.
 
     Examples
@@ -492,21 +540,27 @@ def get_attribute(locator, attribute, anchor='1', element_type=None, timeout=0,
         ${attribute_value}  GetAttribute          Country   value           element_type=dropdown
         ${attribute_value}  GetAttribute          Gender    checked         element_type=checkbox
 
+    Any element using css selectors:
+
+    .. code-block:: robotframework
+
+        ${attribute_value}  GetAttribute          input[type=button]   value    element_type=css
+
     All flags are available for using (timeout, anchor, index, visibility, parent, child etc.).
-    in same way as you are using those with Pacewords like ClickText/Item, TypeText, Dropdown etc.
+    in same way as you are using those with Qwords like ClickText/Item, TypeText, Dropdown etc.
 
     Parameters
     ----------
     locator : str
         Visible text, attribute value or Xpath expression with or without xpath= prefix.
-        The equal sign "=" must be escaped with a "\".
+        The equal sign "=" must be escaped with a "\\".
     attribute: str
         Attribute which value we want to get.
     anchor : int
         Used when element_type is defined. Default=1 (first match)
     element_type : string
         Define element type/preferred searching method
-        (available types: text, input, checkbox, item, dropdown).
+        (available types: text, input, checkbox, item, dropdown or css).
     timeout : int
         How long we wait element to appear. Default=10 sec
     kwargs :
@@ -526,8 +580,8 @@ def get_attribute(locator, attribute, anchor='1', element_type=None, timeout=0,
     webelement = get_webelement(locator, anchor, element_type, timeout, **kwargs)
 
     if not webelement:
-        raise QWebElementNotFoundError('Could not find element {} with attribute {}'
-                                       .format(locator, attribute))
+        raise QWebElementNotFoundError('Could not find element {} with attribute {}'.format(
+            locator, attribute))
     if not isinstance(webelement, list):
         return webelement.get_attribute(attribute)
 
@@ -535,12 +589,19 @@ def get_attribute(locator, attribute, anchor='1', element_type=None, timeout=0,
         return webelement[0].get_attribute(attribute)
 
     raise QWebValueError('Found {} occurences of locator {}. '
-                         'Use index etc. to uniquely identify the element'
-                         .format(len(webelement), locator))
+                         'Use index etc. to uniquely identify the element'.format(
+                             len(webelement), locator))
 
 
+@keyword(tags=["Verification"])
 @decorators.timeout_decorator
-def verify_attribute(locator, attribute, value, anchor='1', element_type=None, timeout=0, **kwargs):
+def verify_attribute(locator: str,
+                     attribute: str,
+                     value: str,
+                     anchor: str = '1',
+                     element_type: Optional[str] = None,
+                     timeout: Union[int, float, str] = 0,
+                     **kwargs) -> None:
     r"""Verify attribute value of an element.
 
     Examples
@@ -572,14 +633,40 @@ def verify_attribute(locator, attribute, value, anchor='1', element_type=None, t
         VerifyAttribute     Country   value     Finland         element_type=dropdown
         VerifyAttribute     Gender    checked   checked         element_type=checkbox
 
+    Any element using css selectors:
+
+    .. code-block:: robotframework
+
+        VerifyAttribute          input[type=button]   value    Click Me     element_type=css
+
     All flags are available for using (timeout, anchor, index, visibility, parent, child etc.).
     in same way as you are using those with Pacewords like ClickText/Item, TypeText, Dropdown etc.
 
+    In addition, operator performing verification can be given as keyword argument.
+    If 'operator' argument is not given, 'equals' is expected. Examples:
+
+    .. code-block:: robotframework
+
+        # Not equal
+        VerifyAttribute     //img    data-icon    screen123    operator=not equal
+        VerifyAttribute     //img    data-icon    screen123    operator=!=
+
+        # Greater than
+        VerifyAttribute    Button3    data-id    7    element_type=Text    operator=greater than
+        VerifyAttribute    Button3    data-id    5    element_type=Text    operator=>
+
+        # less than or equal
+        # ('operator=less than or equal' also works)
+        VerifyAttribute    Button3    data-id    10   element_type=Item   operator=<=
+
+        # contains
+        VerifyAttribute    somebutton    id    skim      element_type=Text   operator=contains
+
     Parameters
     ----------
     locator : str
         Visible text, attribute value or Xpath expression with or without xpath= prefix.
-        The equal sign "=" must be escaped with a "\".
+        The equal sign "=" must be escaped with a "\\".
     attribute: str
         Attribute which value we want to get.
     value: str
@@ -588,7 +675,7 @@ def verify_attribute(locator, attribute, value, anchor='1', element_type=None, t
         Used when element_type is defined. Default=1 (first match)
     element_type : string
         Define element type/preferred searching method
-        (available types: text, input, checkbox, item, dropdown).
+        (available types: text, input, checkbox, item, dropdown or css).
     timeout : int
         How long we wait element to appear. Default=10 sec
     kwargs :
@@ -596,13 +683,28 @@ def verify_attribute(locator, attribute, value, anchor='1', element_type=None, t
         |       Any available for picked searching method.
         |       See interacting with text, item, input etc. elements from
         |       documentation
+        |
+        |       In addition:
+        |       operator : str
+        |           Which operator use for verification. Possible values:
+        |               * 'Equals' or '==' (Default)
+        |               * 'Not Equal' or '!='
+        |               * 'Less than' or '<'
+        |               * 'Greater than' or '>'
+        |               * 'Less than or equal' or  '<='
+        |               * 'Greater than or equal' or '>='
+        |               * 'Contains'
+        |               * 'Not Contains'
+
 
     Related keywords
     ----------------
     \`GetAttribute\`, \`VerifyElement\`
     """
     attr_val = get_attribute(locator, attribute, anchor, element_type, timeout, **kwargs)
+    operator = kwargs.get('operator', "equals")
 
-    if attr_val != value:
-        raise QWebValueError("Expected attribute value differs from real value: {}/{}"
-                             .format(value, attr_val))
+    try:
+        element.operator_verify(attr_val, value, operator)
+    except QWebValueError as e:
+        raise e
