@@ -14,20 +14,25 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 # ---------------------------
-
 """Keywords for input elements.
 
 Input elements are those in which one can input text in.
 """
+from __future__ import annotations
+from typing import Union
+from selenium.webdriver.common.action_chains import ActionChains
+from selenium.webdriver.remote.webelement import WebElement
 from robot.api import logger
+from robot.api.deco import keyword
 from QWeb.internal.exceptions import QWebFileNotFoundError, QWebValueError
-from QWeb.internal import secrets, actions, util
+from QWeb.internal import javascript, secrets, actions, util
 from QWeb.internal import element, input_, download, decorators
 from QWeb.internal.input_handler import INPUT_HANDLER as input_handler
-from selenium.webdriver.remote.webelement import WebElement
+from QWeb.keywords import browser
 
 
-def set_input_handler(input_method):
+@keyword(tags=("Config", "Input"))
+def set_input_handler(input_method: str) -> None:
     """*DEPRECATED!!* Use keyword `SetConfig` instead.
 
     Set input handler.
@@ -58,7 +63,8 @@ def set_input_handler(input_method):
     input_handler.input_method = input_method
 
 
-def set_line_break(key):
+@keyword(tags=("Config", "Input"))
+def set_line_break(key: str) -> str:
     r"""*DEPRECATED!!* Use keyword `SetConfig` instead."""
     old_line_break_key = input_handler.line_break_key
     input_handler.line_break_key = key
@@ -68,7 +74,12 @@ def set_line_break(key):
 secrets.add_filter("Type Secret3", 1, "hint")
 
 
-def type_secret3(locator, input_text, anchor="1", timeout=0, index=1, **kwargs):
+def type_secret3(locator: str,
+                 input_text: str,
+                 anchor: str = "1",
+                 timeout: Union[int, float, str] = 0,
+                 index: int = 1,
+                 **kwargs) -> None:
     """Type sensitive information such as personal data.
 
     the first 1 or 3 digits of the secret value are displayed
@@ -101,7 +112,13 @@ def type_secret3(locator, input_text, anchor="1", timeout=0, index=1, **kwargs):
 secrets.add_filter("Type Secret", 1, None)
 
 
-def type_secret(locator, input_text, anchor="1", timeout=0, index=1, **kwargs):
+@keyword(tags=("Input", "Interaction"))
+def type_secret(locator: str,
+                input_text: str,
+                anchor: str = "1",
+                timeout: Union[int, float, str] = 0,
+                index: int = 1,
+                **kwargs) -> None:
     r"""Type secret information such as password.
 
     Logging in start_keyword and end_keyword is filtered,
@@ -129,8 +146,14 @@ def type_secret(locator, input_text, anchor="1", timeout=0, index=1, **kwargs):
     type_text(locator, input_text, anchor, timeout=timeout, index=index, **kwargs)
 
 
+@keyword(tags=("Input", "Interaction"))
 @decorators.timeout_decorator
-def type_text(locator, input_text, anchor=u"1", timeout=0, index=1, **kwargs):
+def type_text(locator: Union[WebElement, str],
+              input_text: str,
+              anchor: str = "1",
+              timeout: Union[int, float, str] = 0,
+              index: int = 1,
+              **kwargs) -> None:
     r"""Type given text to a text field.
 
     First look through if there are any input fields that have the
@@ -153,7 +176,7 @@ def type_text(locator, input_text, anchor=u"1", timeout=0, index=1, **kwargs):
         Text that locates the input field. The input field that is closest
         to the text is selected. Also one can use xpath by adding xpath= prefix
         and then the xpath. Error is raised if the xpath matches to multiple
-        elements. When using XPaths, the equal sign "=" must be escaped with a "\".
+        elements. When using XPaths, the equal sign "=" must be escaped with a "\\".
         Can also be a WebElement instance returned by GetWebElement keyword or javascript.
     input_text : str
         Text that will be written in the input field
@@ -234,6 +257,9 @@ def type_text(locator, input_text, anchor=u"1", timeout=0, index=1, **kwargs):
         # using WebElement instance
         ${elem}=            GetWebElement  //input[@title\="Search"]
         TypeText            ${elem}     Text to search for
+        # Type to input in shadow DOM input
+        SetConfig           ShadowDOM   True
+        TypeText            Search      Test
 
     Related keywords
     ----------------
@@ -242,13 +268,18 @@ def type_text(locator, input_text, anchor=u"1", timeout=0, index=1, **kwargs):
     if isinstance(locator, WebElement):
         input_element = locator
     else:
-        input_element = input_.get_input_elements_from_all_documents(
-            locator, anchor, timeout=timeout, index=index, **kwargs)
+        input_element = input_.get_input_elements_from_all_documents(locator,
+                                                                     anchor,
+                                                                     timeout=timeout,
+                                                                     index=index,
+                                                                     **kwargs)
     actions.write(input_element, str(input_text), timeout=timeout, **kwargs)
 
 
+@keyword(tags=("File", "Input", "Interaction"))
 @decorators.timeout_decorator
-def type_texts(input_texts, timeout='0'):
+def type_texts(input_texts: Union[dict[str, str], str],
+               timeout: Union[int, float, str] = '0') -> None:
     r"""Type text to multiple fields.
 
     Accepts a .txt file or Robot FW dictionary as a parameter. If using a text file, the locator
@@ -277,17 +308,23 @@ def type_texts(input_texts, timeout='0'):
         file = download.get_path(input_texts)
         with open(file, 'rb') as txt_file:
             params = [line.rstrip() for line in txt_file]
-            for x in params:
-                x = x.decode('utf-8').split(',')
-                locator, text = x[0].strip(), x[1].strip()
+            for x_bytes in params:
+                x_str_list = x_bytes.decode('utf-8').split(',')
+                locator, text = x_str_list[0].strip(), x_str_list[1].strip()
                 logger.info('Typing "{}", locator "{}"'.format(text, locator))
                 type_text(locator, text, timeout=timeout)
     else:
         raise QWebValueError('Unknown input value. Text file or dictionary required.')
 
 
+@keyword(tags=("Input", "Verification"))
 @decorators.timeout_decorator
-def verify_input_value(locator, expected_value, anchor=u"1", timeout=0, index=1, **kwargs):
+def verify_input_value(locator: Union[WebElement, str],
+                       expected_value: str,
+                       anchor: str = "1",
+                       timeout: Union[int, float, str] = 0,
+                       index: int = 1,
+                       **kwargs) -> None:
     r"""Verify input field has given value.
 
     Examples
@@ -311,7 +348,7 @@ def verify_input_value(locator, expected_value, anchor=u"1", timeout=0, index=1,
         Text that locates the input field. The input field that is closest
         to the text is selected. Also one can use xpath by adding xpath= prefix
         and then the xpath. Error is raised if the xpath matches to multiple
-        elements. When using XPaths, the equal sign "=" must be escaped with a "\".
+        elements. When using XPaths, the equal sign "=" must be escaped with a "\\".
     expected_value : str
         Text that the input field should contain.
     anchor : str
@@ -344,13 +381,18 @@ def verify_input_value(locator, expected_value, anchor=u"1", timeout=0, index=1,
     if isinstance(locator, WebElement):
         input_element = locator
     else:
-        input_element = input_.get_input_elements_from_all_documents(
-            locator, anchor, timeout=timeout, index=index, **kwargs)
+        input_element = input_.get_input_elements_from_all_documents(locator,
+                                                                     anchor,
+                                                                     timeout=timeout,
+                                                                     index=index,
+                                                                     **kwargs)
     actions.compare_input_values(input_element, expected_value, timeout=timeout)
 
 
+@keyword(tags=("Input", "Verification"))
 @decorators.timeout_decorator
-def verify_input_values(input_values, timeout='0'):
+def verify_input_values(input_values: Union[dict[str, str], str],
+                        timeout: Union[int, float, str] = '0') -> None:
     r"""Verify input fields have given values.
 
     Accepts a .txt file or Robot FW dictionary as a parameter. If using a text file, the locator
@@ -380,9 +422,9 @@ def verify_input_values(input_values, timeout='0'):
         file = download.get_path(input_values)
         with open(file, 'rb') as txt_file:
             params = [line.rstrip() for line in txt_file]
-            for x in params:
-                x = x.decode('utf-8').split(',')
-                locator, value = x[0].strip(), x[1].strip()
+            for x_bytes in params:
+                x_str_list = x_bytes.decode('utf-8').split(',')
+                locator, value = x_str_list[0].strip(), x_str_list[1].strip()
                 logger.info('Locator: {}, Expected value: {}'.format(locator, value),
                             also_console=True)
                 verify_input_value(locator, value, timeout=timeout)
@@ -390,8 +432,14 @@ def verify_input_values(input_values, timeout='0'):
         raise QWebValueError('Unknown input value. Text file or dictionary required.')
 
 
+@keyword(tags=("Input", "Verification"))
 @decorators.timeout_decorator
-def verify_input_status(locator, status, anchor="1", timeout=0, index=1, **kwargs):
+def verify_input_status(locator: str,
+                        status: str,
+                        anchor: str = "1",
+                        timeout: Union[int, float, str] = 0,
+                        index: int = 1,
+                        **kwargs) -> None:
     r"""Verify input field is enabled or disabled.
 
     In other words verify can user interact with an input field or not.
@@ -419,7 +467,7 @@ def verify_input_status(locator, status, anchor="1", timeout=0, index=1, **kwarg
         Text that locates the input field. The input field that is closest
         to the text is selected. Also one can use xpath by adding xpath= prefix
         and then the xpath. Error is raised if the xpath matches to multiple
-        elements. When using XPaths, the equal sign "=" must be escaped with a "\".
+        elements. When using XPaths, the equal sign "=" must be escaped with a "\\".
     status : str
         Status for the input field. Either enabled, readonly or disabled.
     anchor : str
@@ -449,8 +497,12 @@ def verify_input_status(locator, status, anchor="1", timeout=0, index=1, **kwarg
     ----------------
     \`GetInputValue\`, \`VerifyInputElement\`, \`VerifyInputValue\`, \`VerifyInputValues\`
     """
-    input_element = input_.get_input_elements_from_all_documents(
-        locator, anchor, timeout=timeout, index=index, enable_check=True, **kwargs)
+    input_element = input_.get_input_elements_from_all_documents(locator,
+                                                                 anchor,
+                                                                 timeout=timeout,
+                                                                 index=index,
+                                                                 enable_check=True,
+                                                                 **kwargs)
     if status.lower() == "enabled":
         if not element.is_enabled(input_element) or element.is_readonly(input_element):
             raise QWebValueError('The input field was disabled')
@@ -464,8 +516,13 @@ def verify_input_status(locator, status, anchor="1", timeout=0, index=1, **kwarg
         raise QWebValueError('Unkown status: "{}"'.format(status))
 
 
+@keyword(tags=("Input", "Verification"))
 @decorators.timeout_decorator
-def verify_input_element(locator, anchor='1', timeout=0, index=1, **kwargs):
+def verify_input_element(locator: str,
+                         anchor: str = "1",
+                         timeout: Union[int, float, str] = 0,
+                         index: int = 1,
+                         **kwargs) -> None:
     r"""Verify that input element exist.
 
     Examples
@@ -515,14 +572,22 @@ def verify_input_element(locator, anchor='1', timeout=0, index=1, **kwargs):
     ----------------
     \`GetInputValue\`, \`VerifyInputStatus\`, \`VerifyInputValue\`, \`VerifyInputValues\`
     """
-    input_element = input_.get_input_elements_from_all_documents(
-        locator, anchor, timeout=timeout, index=index, **kwargs)
+    input_element = input_.get_input_elements_from_all_documents(locator,
+                                                                 anchor,
+                                                                 timeout=timeout,
+                                                                 index=index,
+                                                                 **kwargs)
     if input_element:
         return
 
 
+@keyword(tags=("Input", "Getters"))
 @decorators.timeout_decorator
-def get_input_value(locator, anchor='1', timeout=0, index=1, **kwargs):
+def get_input_value(locator: str,
+                    anchor: str = "1",
+                    timeout: Union[int, float, str] = 0,
+                    index: int = 1,
+                    **kwargs) -> Union[int, float, str]:
     r"""Get input value from input field.
 
     Examples
@@ -548,7 +613,7 @@ def get_input_value(locator, anchor='1', timeout=0, index=1, **kwargs):
         Text that locates the input field. The input field that is closest
         to the text is selected. Also one can use xpath by adding xpath= prefix
         and then the xpath. Error is raised if the xpath matches to multiple
-        elements. When using XPaths, the equal sign "=" must be escaped with a "\".
+        elements. When using XPaths, the equal sign "=" must be escaped with a "\\".
     anchor : str
         Index number or text near the input field's locator element.
         If the page contains many places where the locator is then anchor is used
@@ -579,14 +644,23 @@ def get_input_value(locator, anchor='1', timeout=0, index=1, **kwargs):
     ----------------
     \`VerifyInputElement\`, \`VerifyInputStatus\`, \`VerifyInputValue\`, \`VerifyInputValues\`
     """
-    input_element = input_.get_input_elements_from_all_documents(
-        locator, anchor, timeout=timeout, index=index, **kwargs)
+    input_element = input_.get_input_elements_from_all_documents(locator,
+                                                                 anchor,
+                                                                 timeout=timeout,
+                                                                 index=index,
+                                                                 **kwargs)
     val = actions.input_value(input_element, timeout=timeout, **kwargs)
     return util.get_substring(val, **kwargs)
 
 
+@keyword(tags=("File", "Input", "Interaction"))
 @decorators.timeout_decorator
-def upload_file(locator, filename, anchor='1', timeout=0, index=1, **kwargs):
+def upload_file(locator: str,
+                filename: str,
+                anchor: str = "1",
+                timeout: Union[int, float, str] = 0,
+                index: int = 1,
+                **kwargs) -> None:
     r"""Upload file.
 
     Examples
@@ -637,17 +711,24 @@ def upload_file(locator, filename, anchor='1', timeout=0, index=1, **kwargs):
     kwargs['upload'] = util.par2bool(kwargs.get('upload', True))
     filepath = download.get_path(filename)
     if filepath:
-        input_element = input_.get_input_elements_from_all_documents(
-            locator, anchor, timeout=timeout, index=index, **kwargs)
+        input_element = input_.get_input_elements_from_all_documents(locator,
+                                                                     anchor,
+                                                                     timeout=timeout,
+                                                                     index=index,
+                                                                     **kwargs)
         input_element.send_keys(str(filepath.resolve()))
         return
     raise QWebFileNotFoundError(
-        'Unable to find file {}. Tried from project/files and users/downloads'
-        .format(filename))
+        'Unable to find file {}. Tried from project/files and users/downloads'.format(filename))
 
 
+@keyword(tags=("Input", "Interaction"))
 @decorators.timeout_decorator
-def press_key(locator, key, anchor=u"1", timeout='0', **kwargs):
+def press_key(locator: str,
+              key: str,
+              anchor: str = "1",
+              timeout: Union[int, float, str] = 0,
+              **kwargs) -> None:
     r"""Simulate user pressing keyboard key on element identified by "locator".
 
     The parameter "key" is either a single character or a keyboard key surrounded by '{ }'.
@@ -667,10 +748,26 @@ def press_key(locator, key, anchor=u"1", timeout='0', **kwargs):
     ----------------
     \`TypeSecret\`, \`TypeText\`, \`WriteText\`
     """
+    driver = browser.return_browser()
+    action = ActionChains(driver)
     try:
-        input_element = input_.get_input_elements_from_all_documents(
-            locator, anchor, timeout=timeout, index=1, **kwargs)
-        key = input_handler.check_key(key)
-        input_element.send_keys(key)
-    except AttributeError:
-        raise QWebValueError('Could not find key "{}"'.format(key))
+        input_element = input_.get_input_elements_from_all_documents(locator,
+                                                                     anchor,
+                                                                     timeout=timeout,
+                                                                     index=1,
+                                                                     **kwargs)
+        key = input_handler.check_key(key)  # type: ignore[assignment]
+
+        # COMMAND key workaround on safari
+        # supports normal text field CMD operations only
+        # (i.e. CMD+C, CMD+A, CMD+V etc.)
+        if len(key) == 2 and key[0] == '\ue03d' and util.is_safari():
+            javascript.execute_javascript("arguments[0].focus();", input_element)
+            action.key_down(key[0]) \
+                  .send_keys(key[1]) \
+                  .key_up(key[0]).perform()
+        else:
+            input_element.send_keys(key)
+    except AttributeError as e:
+        logger.console(e)
+        raise QWebValueError('Could not find key "{}"'.format(key)) from e
