@@ -16,27 +16,37 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 # ---------------------------
-
 """Keywords for text elements.
 
 Text elements are considered to be elements that have visible text. Input
 elements are not considered text elements even though they might have visible
 text.
 """
+from __future__ import annotations
+from typing import Union, Optional
 from pynput.keyboard import Controller
 import pyperclip
-from QWeb.internal.actions import scroll, execute_click_and_verify_condition, hover_to, \
-    text_appearance, scroll_dynamic_web_page, scroll_first_scrollable_parent_element
+from QWeb.internal.actions import scroll as _scroll, \
+    execute_click_and_verify_condition as _execute_click_and_verify_condition, \
+    hover_to as _hover_to, text_appearance as _text_appearance, \
+    scroll_dynamic_web_page as _scroll_dynamic_web_page, \
+    scroll_first_scrollable_parent_element as _scroll_first_scrollable_parent_element
 from QWeb.internal import element, decorators, util, download, text as internal_text
 from QWeb.internal.config_defaults import CONFIG
 from QWeb.internal.exceptions import QWebValueError, QWebEnvironmentError, QWebTimeoutError,\
-    QWebElementNotFoundError, QWebDriverError
+    QWebElementNotFoundError, QWebDriverError, QWebInstanceDoesNotExistError
 from robot.api import logger
+from robot.api.deco import keyword
 import os
 
 
+@keyword(tags=("Text", "Verification"))
 @decorators.timeout_decorator
-def verify_text(text, timeout=0, anchor="1", **kwargs):  # pylint: disable=unused-argument
+def verify_text(
+        text: str,
+        timeout: Union[int, float, str] = 0,  # pylint: disable=unused-argument
+        anchor: str = "1",
+        **kwargs) -> None:
     r"""Verify page contains given text.
 
     Keyword waits until timeout has passed. If timeout is not specified, it
@@ -66,6 +76,15 @@ def verify_text(text, timeout=0, anchor="1", **kwargs):  # pylint: disable=unuse
        DefaultTimeout    2m
        VerifyText        Canis      # Waits 2 minutes
 
+    To enable searching items also from Shadow DOM, use SetConfig Shadow DOM.
+
+    Shadow DOM Example
+    ------------------
+    .. code-block:: robotframework
+
+        SetConfig         Shadow DOM     True
+        VerifyText        Canis
+
     Parameters
     ----------
     text : str
@@ -104,8 +123,9 @@ def verify_text(text, timeout=0, anchor="1", **kwargs):  # pylint: disable=unuse
         return
 
 
+@keyword(tags=("Text", "Verification"))
 @decorators.timeout_decorator
-def verify_no_text(text, timeout=0, **kwargs):  # pylint: disable=unused-argument
+def verify_no_text(text: str, timeout: Union[int, float, str] = 0, **kwargs) -> None:  # pylint: disable=unused-argument
     r"""Verify that page does not contain given text.
 
     Keyword waits until timeout has passed. If timeout is not specified, it
@@ -160,16 +180,21 @@ def verify_no_text(text, timeout=0, **kwargs):  # pylint: disable=unused-argumen
     \`VerifyText\`
     """
     kwargs['css'] = False
-    web_elements = internal_text.get_element_by_locator_text(
-        text, allow_non_existent=True, **kwargs)
+    web_elements = internal_text.get_element_by_locator_text(text,
+                                                             allow_non_existent=True,
+                                                             **kwargs)
     if not web_elements:
         return
-    raise QWebValueError('Page contained the text "{}" after timeout'.format(
-        text))
+    raise QWebValueError('Page contained the text "{}" after timeout'.format(text))
 
 
+@keyword(tags=("Text", "Verification"))
 @decorators.timeout_decorator
-def verify_text_count(text, expected_count, timeout=0, **kwargs):  # pylint: disable=unused-argument
+def verify_text_count(
+        text: str,
+        expected_count: Union[int, str],
+        timeout: Union[int, float, str] = 0,  # pylint: disable=unused-argument
+        **kwargs) -> None:
     r"""Verify page contains given text given times.
 
     Keyword waits until timeout has passed. If timeout is not specified, it
@@ -235,12 +260,13 @@ def verify_text_count(text, expected_count, timeout=0, **kwargs):  # pylint: dis
     if element_count == expected_count:
         return
 
-    raise QWebValueError('Page contained {0} texts instead of {1} after timeout'
-                         .format(element_count, expected_count))
+    raise QWebValueError('Page contained {0} texts instead of {1} after timeout'.format(
+        element_count, expected_count))
 
 
+@keyword(tags=("Text", "Getters"))
 @decorators.timeout_decorator
-def get_text_count(text, timeout=0, **kwargs):  # pylint: disable=unused-argument
+def get_text_count(text: str, timeout: Union[int, float, str] = 0, **kwargs) -> int:  # pylint: disable=unused-argument
     r"""Get count of appearances for given text.
 
     Keyword waits until timeout has passed. If timeout is not specified, it
@@ -283,9 +309,15 @@ def get_text_count(text, timeout=0, **kwargs):  # pylint: disable=unused-argumen
     return len(web_elements)
 
 
+@keyword(tags=("Text", "Interaction"))
 @decorators.timeout_decorator
-def click_text(text, anchor="1", timeout=0, parent=None,
-               child=None, js=None, **kwargs):
+def click_text(text: str,
+               anchor: str = "1",
+               timeout: Union[int, float, str] = 0,
+               parent: Optional[str] = None,
+               child: Optional[str] = None,
+               js: bool = False,
+               **kwargs) -> None:
     r"""Click text on web page.
 
     Keyword looks for an exact match and if not found uses xpath defined in
@@ -342,6 +374,7 @@ def click_text(text, anchor="1", timeout=0, parent=None,
         ClickText   Canis     js=true     #Use Javascript click instead of Selenium
 
     To double-click text, use argument doubleclick=True
+
     .. code-block:: robotframework
 
         ClickText   Canis       doubleclick=True
@@ -353,6 +386,15 @@ def click_text(text, anchor="1", timeout=0, parent=None,
         SetConfig   DoubleClick     On
         ClickText   Canis
 
+    To search text also from shadow DOM, use SetConfig ShadowDOM.
+    Note that certain arguments like partial_match are ignored when
+    searching from shadow DOM.
+
+    .. code-block:: robotframework
+
+        SetConfig   ShadowDOM     True
+        ClickText   Canis
+
     Parameters
     ----------
     text : str
@@ -378,14 +420,26 @@ def click_text(text, anchor="1", timeout=0, parent=None,
     \`ClickElement\`, \`ClickItem\`, \`GetText\`, \`RightClick\`, \`VerifyText\`
     """
     anchor = str(anchor)
-    web_element = internal_text.get_element_by_locator_text(
-        text, anchor, parent=parent, child=child, **kwargs)
-    if execute_click_and_verify_condition(web_element, timeout=timeout, js=js, **kwargs):
+
+    web_element = internal_text.get_element_by_locator_text(text,
+                                                            anchor,
+                                                            parent=parent,
+                                                            child=child,
+                                                            **kwargs)
+    if _execute_click_and_verify_condition(web_element, timeout=timeout, js=js, **kwargs):
         return
 
 
-def scan_click(text, text_to_appear, anchor="1", timeout=0, interval=None,
-               parent=None, child=None, js=False, **kwargs):  # pylint: disable=unused-argument
+def scan_click(
+        text: str,
+        text_to_appear: str,
+        anchor: str = "1",
+        timeout: Union[int, float, str] = 0,
+        interval: Optional[str] = None,
+        parent: Optional[str] = None,  # pylint: disable=unused-argument
+        child: Optional[str] = None,  # pylint: disable=unused-argument
+        js: bool = False,  # pylint: disable=unused-argument
+        **kwargs) -> None:
     """*DEPRECATED!!* Use keyword `ClickUntil` instead.
 
     Click text until a text appears.
@@ -431,12 +485,26 @@ def scan_click(text, text_to_appear, anchor="1", timeout=0, interval=None,
     if el_type.lower() != 'text':
         click_item_until(text_to_appear, text, anchor, timeout, interval, js=False, **kwargs)
     else:
-        click_until(text_to_appear, text, anchor, timeout,
-                    interval, parent=None, child=None, js=False, **kwargs)
+        click_until(text_to_appear,
+                    text,
+                    anchor,
+                    timeout,
+                    interval,
+                    parent=None,
+                    child=None,
+                    js=False,
+                    **kwargs)
 
 
-def skim_click(text, text_to_disappear='', anchor="1", timeout=0, interval=None,
-               parent=None, child=None, js=False, **kwargs):
+def skim_click(text: str,
+               text_to_disappear: str = '',
+               anchor: str = "1",
+               timeout: Union[int, float, str] = 0,
+               interval: Optional[str] = None,
+               parent: Optional[str] = None,
+               child: Optional[str] = None,
+               js: bool = False,
+               **kwargs) -> None:
     """*DEPRECATED!!* Use keyword `ClickWhile` instead.
 
     Click text until a text disappears.
@@ -492,8 +560,12 @@ def skim_click(text, text_to_disappear='', anchor="1", timeout=0, interval=None,
         click_while(text_to_disappear, text, anchor, timeout, interval, parent, child, js, **kwargs)
 
 
+@keyword(tags=("Text", "Interaction"))
 @decorators.timeout_decorator
-def hover_text(text, anchor="1", timeout="0", **kwargs):  # pylint: disable=unused-argument
+def hover_text(text: str,
+               anchor: str = "1",
+               timeout: Union[int, float, str] = "0",
+               **kwargs) -> None:
     r"""Hover over text.
 
     Keyword will fail if text is not visible. You should use VerifyText before HoverText
@@ -527,12 +599,16 @@ def hover_text(text, anchor="1", timeout="0", **kwargs):  # pylint: disable=unus
         by visible text
     """
     web_element = internal_text.get_element_by_locator_text(text, anchor, **kwargs)
-    hover_to(web_element, timeout=timeout)
+    _hover_to(web_element, timeout=timeout)
 
 
+@keyword(tags=("Item", "Verification"))
 @decorators.timeout_decorator
-def hover_item(locator, anchor="1", timeout="0", **kwargs):  # pylint: disable=unused-argument
-    r"""Hover over text.
+def hover_item(locator: str,
+               anchor: str = "1",
+               timeout: Union[int, float, str] = "0",
+               **kwargs) -> None:
+    r"""Hover over item.
 
     Hover over web element.
 
@@ -568,11 +644,12 @@ def hover_item(locator, anchor="1", timeout="0", **kwargs):  # pylint: disable=u
     \`HoverElement\`, \`HoverText\`, \`HoverTo\`, \`ScrollText\`, \`ScrollTo\`
     """
     web_element = internal_text.get_item_using_anchor(locator, anchor, **kwargs)
-    hover_to(web_element, timeout=timeout)
+    _hover_to(web_element, timeout=timeout)
 
 
+@keyword(tags=("Text", "Verification"))
 @decorators.timeout_decorator
-def is_text(text, timeout="0.1s", **kwargs):
+def is_text(text: str, timeout: Union[int, float, str] = "0.1s", **kwargs) -> bool:
     r"""Return True/False if text is found on the screen.
 
     Used to get text presence to variable. This keyword returns after text is found.
@@ -605,13 +682,14 @@ def is_text(text, timeout="0.1s", **kwargs):
     \`GetText\`, \`IsNoText\`, \`VerifyText\`
     """
     try:
-        return text_appearance(text, text_appear=True, timeout=timeout, **kwargs)
-    except (QWebTimeoutError, QWebValueError):
+        return _text_appearance(text, text_appear=True, timeout=timeout, **kwargs)
+    except (QWebTimeoutError, QWebValueError, QWebInstanceDoesNotExistError):
         return False
 
 
+@keyword(tags=("Text", "Verification"))
 @decorators.timeout_decorator
-def is_no_text(text, timeout="2s", **kwargs):
+def is_no_text(text: str, timeout: Union[int, float, str] = "2s", **kwargs) -> bool:
     r"""Return True/False if text is found on the screen.
 
     Used to get text presence info to variable. This keyword returns if text is not visible or
@@ -645,13 +723,18 @@ def is_no_text(text, timeout="2s", **kwargs):
     \`IsText\`, \`VerifyText\`
     """
     try:
-        return text_appearance(text, text_appear=False, timeout=timeout, **kwargs)
+        return _text_appearance(text, text_appear=False, timeout=timeout, **kwargs)
     except (QWebValueError, QWebTimeoutError):
         return False
 
 
+@keyword(tags=("Text", "Verification"))
 @decorators.timeout_decorator
-def verify_element_text(locator, text_to_find, timeout=0, anchor="1", **kwargs):
+def verify_element_text(locator: str,
+                        text_to_find: str,
+                        timeout: Union[int, float, str] = 0,
+                        anchor: str = "1",
+                        **kwargs) -> None:
     r"""Verify that element contains specified text.
 
     Examples
@@ -665,7 +748,7 @@ def verify_element_text(locator, text_to_find, timeout=0, anchor="1", **kwargs):
     ----------
     locator : str
         Visible text, some attribute of wanted element or Xpath expression without xpath= prefix.
-        When using XPaths, the equal sign "=" must be escaped with a "\".
+        When using XPaths, the equal sign "=" must be escaped with a "\\".
     text_to_find : str
         The text we are trying to verify from the element.
     timeout : int
@@ -691,8 +774,13 @@ def verify_element_text(locator, text_to_find, timeout=0, anchor="1", **kwargs):
     raise QWebValueError(f'"{text_to_find}" not in "{locator_text}"')
 
 
+@keyword(tags=("Text", "Getters"))
 @decorators.timeout_decorator
-def get_text(locator, timeout=0, anchor="1", **kwargs):  # pylint: disable=unused-argument
+def get_text(
+        locator: str,
+        timeout: Union[int, float, str] = 0,  # pylint: disable=unused-argument
+        anchor: str = "1",
+        **kwargs) -> Union[str, int, float]:
     r"""Get text from element specified by xpath.
 
     Examples
@@ -720,7 +808,7 @@ def get_text(locator, timeout=0, anchor="1", **kwargs):  # pylint: disable=unuse
     ----------
     locator : str
         Visible text, some attribute of wanted element or Xpath expression without xpath= prefix.
-        When using XPaths, the equal sign "=" must be escaped with a "\".
+        When using XPaths, the equal sign "=" must be escaped with a "\\".
     timeout : int
         How long we wait that element is found before failing.
     anchor : str
@@ -746,22 +834,29 @@ def get_text(locator, timeout=0, anchor="1", **kwargs):  # pylint: disable=unuse
     tag = kwargs.get('tag', None)
     if tag:
         try:
-            anchor = int(anchor) - 1
+            anchor_int = int(anchor) - 1
             kwargs['element_kw'] = True
-            web_element = element.get_elements_by_attributes(tag, locator, **kwargs)[anchor]
-        except ValueError:
+            web_element = element.get_elements_by_attributes(tag, locator, **kwargs)[anchor_int]
+        except ValueError as e:
             raise QWebValueError(
-                'Only index is allowed anchor when searching element by it\'s attribute')
+                'Only index is allowed anchor when searching element by it\'s attribute') from e
     elif '//' not in locator:
         web_element = internal_text.get_text_using_anchor(locator, anchor)
     else:
-        web_element = element.get_unique_element_by_xpath(locator)
+        web_element = element.get_unique_element_by_xpath(locator, **kwargs)
     text = web_element.text
+    if CONFIG['SearchMode']:
+        element.draw_borders(web_element)
     return util.get_substring(text, **kwargs)
 
 
+@keyword(tags=("Item", "Interaction"))
 @decorators.timeout_decorator
-def click_item(text, anchor="1", timeout=0, js=False, **kwargs):
+def click_item(text: str,
+               anchor: str = "1",
+               timeout: Union[int, float, str] = 0,
+               js: bool = False,
+               **kwargs) -> None:
     r"""Click item (usually icon or picture) on webpage.
 
     Finds webelement by it's tooltip text (title or alt) or some another
@@ -796,6 +891,15 @@ def click_item(text, anchor="1", timeout=0, js=False, **kwargs):
         ClickItem    Canis    3     # clicks the third "Canis" on the page
         ClickItem    Canis    Dog   # clicks the "Canis" near to the word "Dog"
 
+    To enable searching items also from Shadow DOM, use SetConfig   Shadow DOM.
+
+    Shadow DOM Example
+    ------------------
+    .. code-block:: robotframework
+
+        SetConfig        Shadow DOM     True
+        ClickItem        Canis
+
     Parameters
     ----------
     text : str
@@ -822,12 +926,17 @@ def click_item(text, anchor="1", timeout=0, js=False, **kwargs):
     \`ClickList\`, \`ClickText\`, \`ClickUntil\`, \`ClickWhile\`, \`VerifyItem\`
     """
     web_element = internal_text.get_item_using_anchor(text, anchor, **kwargs)
-    if execute_click_and_verify_condition(web_element, timeout=timeout, js=js, **kwargs):
+    if _execute_click_and_verify_condition(web_element, timeout=timeout, js=js, **kwargs):
         return
 
 
+@keyword(tags=("Item", "Verification"))
 @decorators.timeout_decorator
-def verify_item(text, anchor="1", timeout=0, **kwargs):  # pylint: disable=unused-argument
+def verify_item(
+        text: str,
+        anchor: str = "1",
+        timeout: Union[int, float, str] = 0,  # pylint: disable=unused-argument
+        **kwargs) -> None:
     r"""Verify Item (usually icon or picture) exist.
 
     Finds webelement by it's tooltip text (title or alt) or some another
@@ -856,6 +965,16 @@ def verify_item(text, anchor="1", timeout=0, **kwargs):  # pylint: disable=unuse
     - a number or
     - a word that is near to the word Canis
 
+    To enable searching items also from Shadow DOM, use SetConfig Shadow DOM.
+
+    Shadow DOM Example
+    ------------------
+    .. code-block:: robotframework
+
+        SetConfig         Shadow DOM     True
+        VerifyItem        Canis
+
+
     Parameters
     ----------
     text : str
@@ -883,8 +1002,13 @@ def verify_item(text, anchor="1", timeout=0, **kwargs):  # pylint: disable=unuse
         return
 
 
+@keyword(tags=("Item", "Verification"))
 @decorators.timeout_decorator
-def verify_no_item(text, anchor="1", timeout=0, **kwargs):  # pylint: disable=unused-argument
+def verify_no_item(
+        text: str,
+        anchor: str = "1",
+        timeout: Union[int, float, str] = 0,  # pylint: disable=unused-argument
+        **kwargs) -> None:
     r"""Verify Item (usually icon or picture) is not exist.
 
     Finds webelement by it's tooltip text (title or alt) or some another
@@ -935,8 +1059,120 @@ def verify_no_item(text, anchor="1", timeout=0, **kwargs):  # pylint: disable=un
     raise QWebValueError('Element with attribute value {} still exists'.format(text))
 
 
+@keyword(tags=("Item", "Verification"))
 @decorators.timeout_decorator
-def scroll_text(text, anchor="1", timeout=0, **kwargs):
+def is_item(text: str, anchor: str = "1", timeout: Union[int, float, str] = 0.1, **kwargs) -> bool:
+    r"""Check if Item (usually icon or picture) exists.
+
+    Finds webelement by it's tooltip text (title or alt) or some another
+    attribute.
+    Available attributes: any
+    Available element types without using tag attribute:
+    *a, span, img, li, h1, h2, h3, h4, h5, h6, div, svg, p, button, input\*
+    (\*submit buttons only)*
+
+    Keyword waits until timeout has passed or item is found.
+    If timeout is not specified, it uses default timeout that can be adjusted
+    with DefaultTimeout keyword.
+
+    IsItem does not require for the attribute to be unique.
+
+    Examples
+    --------
+    .. code-block:: robotframework
+
+        ${exists}=  IsItem        Canis
+
+    In the above example the IsItem keyword returns True if element which
+    has attribute Canis exists.
+
+    Parameters
+    ----------
+    text : str
+        Attribute value of item.
+    anchor : str
+        Parameter not used with this kw.
+    timeout : str | int
+        How long we wait for element to be ready for click
+    kwargs :
+        |  Accepted kwargs:
+        |       tag : html tag of preferred element -
+        |           If tag is used then element is found
+        |           by some of it's attribute
+        |       partial_match: False. If only full match is accepted set
+        |       partial_match to False.
+
+    Related keywords
+    ----------------
+    \`ClickItem\`, \`IsNoItem\, \`VerifyItem\`
+    """
+    try:
+        verify_item(text, anchor, timeout, **kwargs)
+        return True
+    except QWebElementNotFoundError:
+        return False
+
+
+@keyword(tags=("Item", "Verification"))
+@decorators.timeout_decorator
+def is_no_item(text: str, anchor: str = "1", timeout: Union[int, float, str] = 0.1, **kwargs):
+    r"""Check if Item (usually icon or picture) does not exist.
+
+    Finds webelement by it's tooltip text (title or alt) or some another
+    attribute.
+    Available attributes: any
+    Available element types without using tag attribute:
+    *a, span, img, li, h1, h2, h3, h4, h5, h6, div, svg, p, button, input\*
+    (\*submit buttons only)*
+
+    Keyword waits until timeout has passed or element disappears from the screen.
+    If timeout is not specified, it uses default timeout that can be adjusted
+    with DefaultTimeout keyword.
+
+    IsNoItem does not require for the attribute to be unique.
+
+    Examples
+    --------
+    .. code-block:: robotframework
+
+        ${exists}=  IsNoItem        Canis
+
+    In the above example the IsNoItem keyword returns True if element which
+    has attribute Canis disappears from the screen before timeout expires.
+
+    Parameters
+    ----------
+    text : str
+        Attribute value of item.
+    anchor : str
+        Parameter not used with this kw.
+    timeout : str | int
+        How long we wait for element to be ready for click
+    kwargs :
+        |  Accepted kwargs:
+        |       tag : html tag of preferred element -
+        |           If tag is used then element is found
+        |           by some of it's attribute
+        |       partial_match: False. If only full match is accepted set
+        |       partial_match to False.
+
+    Related keywords
+    ----------------
+    \`ClickItem\`, \`IsItem\, \`VerifyItem\`
+    """
+    try:
+        verify_no_item(text, anchor, timeout, **kwargs)
+        return True
+    except QWebValueError:
+        return False
+
+
+@keyword(tags=("Text", "Interaction"))
+@decorators.timeout_decorator
+def scroll_text(text: str,
+                anchor: str = "1",
+                timeout: Union[int, float, str] = 0,
+                **kwargs) -> None:
     r"""Scroll page until text is on top.
 
     Finds text on page and scrolls until it is on top. This is used with visual testing
@@ -958,10 +1194,11 @@ def scroll_text(text, anchor="1", timeout=0, **kwargs):
     \`HoverElement\`, \`HoverItem\`, \`HoverText\`, \`HoverTo\`, \`ScrollTo\`
     """
     web_element = internal_text.get_element_by_locator_text(text, anchor, **kwargs)
-    scroll(web_element, timeout=timeout)
+    _scroll(web_element, timeout=timeout)
 
 
-def write_text(text):
+@keyword(tags=("input", "Text", "Interaction"))
+def write_text(text: str) -> None:
     r"""Type text with single-character keys.
 
     Parameters
@@ -980,14 +1217,14 @@ def write_text(text):
     \`PressKey\`, \`TypeText\`
     """
     if CONFIG.get_value("Headless") or os.getenv('QWEB_HEADLESS', None):
-        raise QWebEnvironmentError(
-            'Running in headless environment. Pynput is unavailable.')
+        raise QWebEnvironmentError('Running in headless environment. Pynput is unavailable.')
     keyboard = Controller()
     keyboard.type(text)
 
 
 @decorators.timeout_decorator
-def verify_texts(texts_to_verify, timeout=0):
+def verify_texts(texts_to_verify: Union[list[str], str],
+                 timeout: Union[int, float, str] = 0) -> None:
     r"""*DEPRECATED!!* Use keyword `VerifyAll` instead."""
     logger.warn(r"""*DEPRECATED!!* Use keyword `VerifyAll` instead.""")
     if isinstance(texts_to_verify, list):
@@ -999,9 +1236,9 @@ def verify_texts(texts_to_verify, timeout=0):
         with open(file, 'rb') as txt_file:
             clean_data = [line.rstrip() for line in txt_file]
             for x in clean_data:
-                x = x.decode('utf-8')
-                logger.info('Verifying text "{}".'.format(x), also_console=True)
-                verify_text(x, timeout)
+                x_str = x.decode('utf-8')
+                logger.info('Verifying text "{}".'.format(x_str), also_console=True)
+                verify_text(x_str, timeout)
     else:
         texts = texts_to_verify.split(',')
         for text in texts:
@@ -1010,8 +1247,9 @@ def verify_texts(texts_to_verify, timeout=0):
             verify_text(clean_text, timeout)
 
 
+@keyword(tags=("File", "Text", "Verification"))
 @decorators.timeout_decorator
-def verify_any(texts_to_verify, timeout=0):
+def verify_any(texts_to_verify: Union[list[str], str], timeout: Union[int, float, str] = 0) -> str:
     r"""Verify any of the given texts.
 
     Verify that at least one of the texts is found. Useful for handling
@@ -1056,9 +1294,9 @@ def verify_any(texts_to_verify, timeout=0):
         with open(file, 'rb') as txt_file:
             clean_data = [line.rstrip() for line in txt_file]
             for x in clean_data:
-                x = x.decode('utf-8')
-                if is_text(x, timeout):
-                    return x
+                x_str = x.decode('utf-8')
+                if is_text(x_str, timeout):
+                    return x_str
 
             raise QWebValueError('Could not find any of the texts: '
                                  f'{clean_data}')
@@ -1075,8 +1313,9 @@ def verify_any(texts_to_verify, timeout=0):
                              f'{texts}')
 
 
+@keyword(tags=("File", "Text", "Verification"))
 @decorators.timeout_decorator
-def verify_all(texts_to_verify, timeout=0):
+def verify_all(texts_to_verify: Union[list[str], str], timeout: Union[int, float, str] = 0) -> None:
     r"""Verify page contains given texts.
 
     The texts should be separated with a comma.
@@ -1109,9 +1348,9 @@ def verify_all(texts_to_verify, timeout=0):
         with open(file, 'rb') as txt_file:
             clean_data = [line.rstrip() for line in txt_file]
             for x in clean_data:
-                x = x.decode('utf-8')
-                logger.info('Verifying text "{}".'.format(x), also_console=True)
-                verify_text(x, timeout)
+                x_str = x.decode('utf-8')
+                logger.info('Verifying text "{}".'.format(x_str), also_console=True)
+                verify_text(x_str, timeout)
     else:
         texts = texts_to_verify.split(',')
         for text in texts:
@@ -1120,9 +1359,14 @@ def verify_all(texts_to_verify, timeout=0):
             verify_text(clean_text, timeout)
 
 
+@keyword(tags=("Text", "Interaction"))
 @decorators.timeout_decorator
-def scroll_to(text_to_find, locator=None, anchor='1', scroll_length=None,
-              timeout=120, **kwargs):  # pylint: disable=unused-argument
+def scroll_to(text_to_find: str,
+              locator: Optional[str] = None,
+              anchor: str = '1',
+              scroll_length: Optional[str] = None,
+              timeout: Union[int, float, str] = 120,
+              **kwargs) -> None:
     r"""Scroll a dynamic web page or scrollbar.
 
     Parameters
@@ -1178,13 +1422,14 @@ def scroll_to(text_to_find, locator=None, anchor='1', scroll_length=None,
         return
     slow_mode = util.par2bool(kwargs.get('slow_mode', False))
     if locator:  # If we are trying to scroll a specific element
-        scroll_first_scrollable_parent_element(
-            locator, anchor, text_to_find, scroll_length, slow_mode, timeout)
+        _scroll_first_scrollable_parent_element(locator, anchor, text_to_find, scroll_length,
+                                                slow_mode, timeout)
     else:  # if we are just scrolling the web page
-        scroll_dynamic_web_page(text_to_find, scroll_length, slow_mode, timeout)
+        _scroll_dynamic_web_page(text_to_find, scroll_length, slow_mode, timeout)
 
 
-def copy_text(text):
+@keyword(tags=("Text", "Interaction"))
+def copy_text(text: str) -> None:
     """Perform a copy.
 
     Copies whatever is active at the UI to the clipboard. If text is
@@ -1206,8 +1451,16 @@ def copy_text(text):
     pyperclip.copy(text)
 
 
-def click_while(text_to_disappear, text_to_click=None, anchor="1", timeout=0, interval=None,
-                parent=None, child=None, js=False, **kwargs):
+@keyword(tags=("Text", "Interaction"))
+def click_while(text_to_disappear: str,
+                text_to_click: Optional[str] = None,
+                anchor: str = "1",
+                timeout: Union[int, float, str] = 0,
+                interval: Optional[str] = None,
+                parent: Optional[str] = None,
+                child: Optional[str] = None,
+                js: bool = False,
+                **kwargs) -> None:
     r"""Click text until a text or element disappears.
 
     This keyword is required on slow pages where it takes time for a button
@@ -1268,21 +1521,29 @@ def click_while(text_to_disappear, text_to_click=None, anchor="1", timeout=0, in
         if el:
             try:
                 verify_item(text_to_disappear, timeout=timeout, **kwargs)
-            except QWebElementNotFoundError:
-                raise QWebValueError('Element to disappear is not visible before click.')
+            except QWebElementNotFoundError as e:
+                raise QWebValueError('Element to disappear is not visible before click.') from e
         elif not is_text(text_to_disappear, timeout):
             raise QWebValueError('Text to disappear is not visible before click.')
     try:
-        _retry_while(text_to_click, text_to_disappear, anchor, timeout, interval,
-                     parent, child, js, **kwargs)
+        _retry_while(text_to_click, text_to_disappear, anchor, timeout, interval, parent, child, js,
+                     **kwargs)
     except QWebElementNotFoundError as orig_err:
         err = CONFIG.get_value('RetryError')
         CONFIG.reset_value('RetryError')
         raise err or orig_err
 
 
-def click_until(text_to_appear, text_to_click, anchor="1", timeout=0, interval=None,
-                parent=None, child=None, js=False, **kwargs):
+@keyword(tags=("Text", "Interaction"))
+def click_until(text_to_appear: str,
+                text_to_click: Optional[str] = None,
+                anchor: str = "1",
+                timeout: Union[int, float, str] = 0,
+                interval: Optional[str] = None,
+                parent: Optional[str] = None,
+                child: Optional[str] = None,
+                js: bool = False,
+                **kwargs) -> None:
     r"""Click text until a text appears.
 
     This keyword is required on slow pages where it takes time for a button
@@ -1342,21 +1603,27 @@ def click_until(text_to_appear, text_to_click, anchor="1", timeout=0, interval=N
         if el:
             try:
                 verify_no_item(text_to_appear, **kwargs)
-            except QWebValueError:
-                raise QWebValueError('Element to appear is already visible.')
+            except QWebValueError as e:
+                raise QWebValueError('Element to appear is already visible.') from e
         elif not is_no_text(text_to_appear):
             raise QWebValueError('Text to appear is already visible.')
     try:
-        _retry_until(text_to_click, text_to_appear, anchor, timeout, interval,
-                     parent, child, js, **kwargs)
+        _retry_until(text_to_click, text_to_appear, anchor, timeout, interval, parent, child, js,
+                     **kwargs)
     except QWebElementNotFoundError as orig_err:
         err = CONFIG.get_value('RetryError')
         CONFIG.reset_value('RetryError')
         raise err or orig_err
 
 
-def click_item_while(text_to_disappear, locator=None, anchor="1", timeout=0,
-                     interval=None, js=False, **kwargs):
+@keyword(tags=("Item", "Interaction"))
+def click_item_while(text_to_disappear: str,
+                     locator: Optional[str] = None,
+                     anchor: str = "1",
+                     timeout: Union[int, float, str] = 0,
+                     interval: Optional[str] = None,
+                     js: bool = False,
+                     **kwargs) -> None:
     r"""Click Item until a text or element disappears.
 
     This keyword is required on slow pages where it takes time for a button
@@ -1415,8 +1682,8 @@ def click_item_while(text_to_disappear, locator=None, anchor="1", timeout=0,
         if el:
             try:
                 verify_item(text_to_disappear, timeout=timeout, **kwargs)
-            except QWebElementNotFoundError:
-                raise QWebValueError('Element to disappear is not visible before click.')
+            except QWebElementNotFoundError as e:
+                raise QWebValueError('Element to disappear is not visible before click.') from e
         elif not is_text(text_to_disappear, timeout):
             raise QWebValueError('Text to disappear is not visible before click.')
     try:
@@ -1427,8 +1694,14 @@ def click_item_while(text_to_disappear, locator=None, anchor="1", timeout=0,
         raise err or orig_err
 
 
-def click_item_until(text_to_appear, locator, anchor="1", timeout=0,
-                     interval=None, js=False, **kwargs):
+@keyword(tags=("Item", "Interaction"))
+def click_item_until(text_to_appear: str,
+                     locator: Optional[str] = None,
+                     anchor: str = "1",
+                     timeout: Union[int, float, str] = 0,
+                     interval: Optional[str] = None,
+                     js: bool = False,
+                     **kwargs) -> None:
     r"""Click text until a text appears.
 
     This keyword is required on slow pages where it takes time for a button
@@ -1483,8 +1756,8 @@ def click_item_until(text_to_appear, locator, anchor="1", timeout=0,
         if el:
             try:
                 verify_no_item(text_to_appear, **kwargs)
-            except QWebValueError:
-                raise QWebValueError('Element to appear is already visible.')
+            except QWebValueError as e:
+                raise QWebValueError('Element to appear is already visible.') from e
         elif not is_no_text(text_to_appear):
             raise QWebValueError('Text to appear is already visible.')
     try:
@@ -1496,8 +1769,16 @@ def click_item_until(text_to_appear, locator, anchor="1", timeout=0,
 
 
 @decorators.timeout_decorator
-def _retry_while(text, text_to_disappear, anchor, timeout, interval,  # pylint: disable=unused-argument
-                 parent=None, child=None, js=False, **kwargs):
+def _retry_while(
+        text: str,
+        text_to_disappear: str,
+        anchor: str,
+        timeout: Union[int, float, str],  # pylint: disable=unused-argument
+        interval: Optional[str],
+        parent: Optional[str] = None,
+        child: Optional[str] = None,
+        js: bool = False,
+        **kwargs) -> None:
     if not interval:
         interval = CONFIG['RetryInterval']
     item = kwargs.get('item', False)
@@ -1522,8 +1803,16 @@ def _retry_while(text, text_to_disappear, anchor, timeout, interval,  # pylint:
 
 
 @decorators.timeout_decorator
-def _retry_until(text, text_to_appear, anchor, timeout, interval,  # pylint: disable=unused-argument
-                 parent=None, child=None, js=False, **kwargs):
+def _retry_until(
+        text: str,
+        text_to_appear: str,
+        anchor: str,
+        timeout: Union[int, float, str],  # pylint: disable=unused-argument
+        interval: Optional[str],
+        parent: Optional[str] = None,
+        child: Optional[str] = None,
+        js: bool = False,
+        **kwargs) -> None:
     if not interval:
         interval = CONFIG['RetryInterval']
     item = kwargs.get('item', False)
